"""

The transformer is a high-level class that hold at least two transformation pipelines

  * One related to the transformation of the input of the model
  * The other related to the target of the model.

It allows accessing the information of the transformed data and is the object that uses the
dataset iterators to transform the data before feeding it to the model.
"""
import copy
import logging
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from sklearn.utils.validation import check_is_fitted

from ceruleo.transformation.functional.graph_utils import topological_sort_iterator
from ceruleo.transformation.functional.pipeline.cache_store import CacheStoreType
from ceruleo.transformation.functional.pipeline.pipeline import Pipeline
from ceruleo.transformation.functional.transformerstep import TransformerStep

logger = logging.getLogger(__name__)


def transformer_info(transformer: Optional[Pipeline]):
    """Obtains the transformer information in a serializable format

    Parameters:

        transformer: The transformer step, or pipeline to obtain their underlying information

    Returns:
        dict

    Raises:

        ValueError
            If the transformer passed as an argument doesn't have
            the get_params method.
    """
    if transformer is None:
        return "Missing"

    data = []
    Q = topological_sort_iterator(transformer)
    for q in Q:
        data.append(q.description())
    return data


class Transformer:
    """Transform each life

    The transformer class is the highest level class of the transformer API.
    It contains Transformation Pipelines for the input data and the target,
    and provides mechanism to inspect the structure of the transformed data.

    Parameters:

        pipelineX: Pipeline that will be applied to the run-to-cycle data
        pipelineY: Pipeline that will be applied to the target.
        pipelineMetadata: Pipeline that will be used to extract additional
                            data from the lives information, by default None
    """

    def __init__(
        self,
        pipelineX: Union[Pipeline, TransformerStep],
        pipelineY: Optional[Union[Pipeline, TransformerStep]] = None,
        pipelineMetadata: Optional[Union[Pipeline, TransformerStep]] = None,
        cache_type: CacheStoreType = CacheStoreType.MEMORY,
    ):
        def ensure_pipeline(x, cache_type: CacheStoreType):
            if isinstance(x, Pipeline):
                return x
            return Pipeline(x, cache_type=cache_type)
        self.cache_type = cache_type
        self.pipelineX = ensure_pipeline(pipelineX, cache_type)
        if pipelineY is not None:
            self.pipelineY = ensure_pipeline(pipelineY, cache_type)
        else:
            self.pipelineY = None
        self.pipelineMetadata = (
            ensure_pipeline(pipelineMetadata, cache_type)
            if pipelineMetadata is not None
            else None
        )
        self.features = None
        self.fitted_ = False

    def _process_selected_features(self):
        if self.pipelineX["selector"] is not None:
            selected_columns = self.pipelineX["selector"].get_support(indices=True)
            self.features = [self.features[i] for i in selected_columns]

    def clone(self):
        return copy.deepcopy(self)

    def fit(self, dataset, show_progress: bool = False):
        """Fit the transformer with a given dataset.

        The transformer will fit the X transformer,
        the Y transformer and the metadata transformer

        Parameters:
            dataset:

        """
        logger.debug("Fitting Transformer")

        self.pipelineX.fit(dataset, show_progress=show_progress)
        if self.pipelineY is not None:
            self.pipelineY.fit(dataset, show_progress=show_progress)
        if self.pipelineMetadata is not None:
            self.pipelineMetadata.fit(dataset)

        if not isinstance(dataset, pd.DataFrame):
            self.minimal_df = dataset[0].head(n=20)
        else:
            self.minimal_df = dataset.head(n=20)
        X = self.pipelineX.transform(self.minimal_df)
        self.number_of_features_ = X.shape[1]
        self.fitted_ = True
        self.column_names = self._compute_column_names()
        return self

    def transform(self, life: pd.DataFrame):
        """Transform a life and obtain the input data, the target and the metadata

        Parameters:
            life: A life in a form of a DataFrame

        Returns:

            Tuple[np.array, np.array, np.array]
                * The first element consists of the input transformed
                * The second element consits of the target transformed
                * The third element consists of the metadata
        """
        check_is_fitted(self, "fitted_")
        return (
            self.transformX(life),
            self.transformY(life),
            self.transformMetadata(life),
        )

    def fit_map(self, dataset, show_progress: bool = False) -> "TransformedDataset":
        self.fit(dataset, show_progress=show_progress)
        return dataset.map(self)

    def transformMetadata(self, df: pd.DataFrame) -> Optional[any]:
        if self.pipelineMetadata is not None:
            return self.pipelineMetadata.transform(df)
        else:
            return None

    def transformY(self, life: pd.DataFrame) -> np.array:
        """Get the transformed target from a life

        Parameters

        life: A run-to-failrue cycle in a form of a DataFrame

        Returns
            t: Target obtained from the life
        """
        if self.pipelineY is not None:
            return self.pipelineY.transform(life)
        else:
            return None

    def transformX(self, life: pd.DataFrame) -> np.array:
        """Get the transformer input data

        Parameters

            life: A life i an form of a DataFrame

        Returns

            t: Input data transformed
        """
        return self.pipelineX.transform(life)

    def columns(self) -> List[str]:
        """Columns names after transformation

        Returns:

            c: columns
        """
        return self.column_names

    @property
    def n_features(self) -> int:
        """Number of features after transformation

        Returns:

            n: Number of features
        """
        return self.number_of_features_

    def _compute_column_names(self):
        return self.pipelineX.column_names

    def description(self):
        return {
            "features": self.features,
            "pipelineX": transformer_info(self.pipelineX),
            "pipelineY": transformer_info(self.pipelineY),
        }

    def __str__(self):
        return str(self.description())

    def get_params(self, deep: bool = False):
        params = {
            "pipelineX": self.pipelineX,
            "pipelineY": self.pipelineY,
            "pipelineMetadata": self.pipelineMetadata,
            "cache_type": self.cache_type,
        }
        if deep:
            paramsX = self.pipelineX.get_params(deep)
            paramsY = self.pipelineY.get_params(deep)
            for k in paramsX.keys():
                params[f"pipeline_X__{k}"] = paramsX[k]
            for k in paramsY.keys():
                params[f"pipeline_Y__{k}"] = paramsY[k]

        return params

    def set_params(self, **params):
        pipeline_X_params = {}
        pipeline_Y_params = {}
        for k in params.keys():
            if k.starts_with("pipeline_X__"):
                new_key = "__".join(k.split("__")[1:])
                pipeline_X_params[new_key] = params[k]
            elif k.starts_with("pipeline_Y__"):
                new_key = "__".join(k.split("__")[1:])
                pipeline_Y_params[new_key] = params[k]
        self.pipelineX = self.pipelineX.set_params(pipeline_X_params)
        self.pipelineY = self.pipelineY.set_params(pipeline_Y_params)
        return self


def TransformerIdentity(rul_column: str = "RUL") -> Transformer:
    """Return the Transformer

    Parameters:

        rul_column : Name of the RUL Column

    Returns:

        TransformerIdentity: An identity f(x)=x transformer
    """
    from ceruleo.transformation.features.selection import ByNameFeatureSelector
    from ceruleo.transformation.utils import IdentityTransformerStep

    return Transformer(
        IdentityTransformerStep(), ByNameFeatureSelector(features=[rul_column])
    )



from ceruleo.dataset.builder.builder import DatasetBuilder
from ceruleo.dataset.builder.cycles_splitter import (
    CyclesSplitter,
    FailureDataCycleSplitter,
    IncreasingFeatureCycleSplitter,
    LifeEndIndicatorCycleSplitter,
    LifeIdCycleSplitter,
)
import pandas as pd

from ceruleo.dataset.builder.output import InMemoryOutputMode, LocalStorageOutputMode
from ceruleo.dataset.builder.rul_column import CycleRULColumn, DatetimeRULColumn
from ceruleo.dataset.ts_dataset import AbstractPDMDataset, PDMDataset
from pathlib import Path 

TEMP_OUTPUT_PATH = Path(__file__).resolve().parent / "temp_output"
TEMP_OUTPUT_DIR_PATH = Path(__file__).resolve().parent / "temp_output_dir"

def is_decreasing(s:pd.Series):
    return all(s[i] >= s[i+1] for i in range(len(s)-1))

def test_dataset_builder_one_file_increasing_feature_cycle_RUL():
    df = pd.DataFrame(
        {
            "Cycle": list(range(0, 12, ))*2,
            "feature_a": list(range(12))*2,
            "feature_b": list(range(12, 24))*2,
        }
    )
    dataset = (
        DatasetBuilder()
        .set_splitting_method(IncreasingFeatureCycleSplitter("Cycle"))
        .set_rul_column_method(CycleRULColumn("Cycle"))
        .set_output_mode(InMemoryOutputMode())
        .build_from_df(df)
    )
    assert isinstance(dataset, AbstractPDMDataset) 
    assert len(dataset) == 2
    assert isinstance(dataset[0], pd.DataFrame)
    assert "RUL" in dataset[0].columns.values.tolist()
    assert is_decreasing(dataset[0].RUL)


def test_dataset_builder_one_file_increasing_feature_datetime_RUL():
    df = pd.DataFrame(
        {
            "Cycle": list(range(0, 12, ))*2,
            "datetime": pd.date_range("2021-01-01", periods=24, freq="min").tolist(),
            "feature_a": list(range(12))*2,
            "feature_b": list(range(12, 24))*2,
        }
    )
    dataset = (
        DatasetBuilder()
        .set_splitting_method(IncreasingFeatureCycleSplitter("Cycle"))
        .set_rul_column_method(DatetimeRULColumn("datetime", "s"))
        .set_output_mode(InMemoryOutputMode())
        .build_from_df(df)
    )
    assert isinstance(dataset, AbstractPDMDataset) 
    assert len(dataset) == 2
    assert isinstance(dataset[0], pd.DataFrame)
    assert "RUL" in dataset[0].columns.values.tolist()
    assert is_decreasing(dataset[0].RUL)
    assert dataset[0].RUL.min() == 0
    assert dataset[0].RUL.max() == 11*60


def test_dataset_builder_one_file_life_id_cycle_datetime_RUL():
    df = pd.DataFrame(
        {
            "life_id": [1]*12 + [2]*12,
            "datetime": pd.date_range("2021-01-01", periods=24, freq="min").tolist(),
            "feature_a": list(range(12))*2,
            "feature_b": list(range(12, 24))*2,
        }
    )
    dataset = (
        DatasetBuilder()
        .set_splitting_method(LifeIdCycleSplitter("life_id"))
        .set_rul_column_method(DatetimeRULColumn("datetime", "s"))
        .set_output_mode(InMemoryOutputMode())
        .build_from_df(df)
    )
    assert isinstance(dataset, AbstractPDMDataset) 
    assert len(dataset) == 2
    assert isinstance(dataset[0], pd.DataFrame)
    assert "RUL" in dataset[0].columns.values.tolist()
    assert is_decreasing(dataset[0].RUL)
    assert dataset[0].RUL.min() == 0
    assert dataset[0].RUL.max() == 11*60


def test_dataset_builder_one_file_life_end_datetime_RUL():
    df = pd.DataFrame(
        {
            "life_end": [0]*11 + [1] + [0]*11 + [1],
            "datetime": pd.date_range("2021-01-01", periods=24, freq="min").tolist(),
            "feature_a": list(range(12))*2,
            "feature_b": list(range(12, 24))*2,
        }
    )
    dataset = (
        DatasetBuilder()
        .set_splitting_method(LifeEndIndicatorCycleSplitter("life_end"))
        .set_rul_column_method(DatetimeRULColumn("datetime", "s"))
        .set_output_mode(InMemoryOutputMode())
        .build_from_df(df)
    )
    assert isinstance(dataset, AbstractPDMDataset) 
    assert len(dataset) == 2
    assert isinstance(dataset[0], pd.DataFrame)
    assert "RUL" in dataset[0].columns.values.tolist()
    assert is_decreasing(dataset[0].RUL)
    assert dataset[0].RUL.min() == 0
    assert dataset[0].RUL.max() == 11*60







def test_dataset_builder_two_file_life_end_datetime_RUL():
    df = pd.DataFrame(
        {
           
            "datetime": pd.date_range("2021-01-01", periods=24, freq="min").tolist(),
            "feature_a": list(range(12))*2,
            "feature_b": list(range(12, 24))*2,
        }
    )
    failures = pd.DataFrame({
        "datetime_failure": [pd.Timestamp("2021-01-01 00:11:00"), pd.Timestamp("2021-01-01 00:23:00")],
        "failure_type": ["A", "B"]
    })
    dataset = (
        DatasetBuilder()
        .set_splitting_method(FailureDataCycleSplitter("datetime", "datetime_failure"))
        .set_rul_column_method(DatetimeRULColumn("datetime", "s"))
        .set_output_mode(InMemoryOutputMode())
        .build_from_data_fault_pair((df, failures))
    )
    assert isinstance(dataset, AbstractPDMDataset) 
    assert len(dataset) == 2
    assert isinstance(dataset[0], pd.DataFrame)
    assert "RUL" in dataset[0].columns.values.tolist()
    assert is_decreasing(dataset[0].RUL)
    assert dataset[0].RUL.min() == 0
    assert dataset[0].RUL.max() == 11*60


def test_dataset_builder_single_file_from_files():
    TEMP_OUTPUT_PATH.mkdir(exist_ok=True, parents=True)
    df1 = pd.DataFrame(
        {
            "Cycle": list(range(0, 12, ))*2,
            "feature_a": list(range(12))*2,
            "feature_b": list(range(12, 24))*2,
        }
    )
    df2 = pd.DataFrame(
        {
            "Cycle": list(range(0, 12, ))*2,
            "feature_a": list(range(12))*2,
            "feature_b": list(range(12, 24))*2,
        }
    )
    df1.to_csv(TEMP_OUTPUT_PATH / "df1.csv", index=False)
    df2.to_csv(TEMP_OUTPUT_PATH / "df2.csv", index=False)
    dataset = (
        DatasetBuilder()
        .set_splitting_method(IncreasingFeatureCycleSplitter("Cycle"))
        .set_rul_column_method(CycleRULColumn("Cycle"))
        .set_output_mode(LocalStorageOutputMode(TEMP_OUTPUT_DIR_PATH))
        .build_from_single_files([TEMP_OUTPUT_PATH / "df1.csv", TEMP_OUTPUT_PATH / "df1.csv"])
    )
    assert isinstance(dataset, AbstractPDMDataset) 
    assert len(dataset) == 4
    assert isinstance(dataset[0], pd.DataFrame)
    assert "RUL" in dataset[0].columns.values.tolist()
    assert is_decreasing(dataset[0].RUL)

    







import numpy as np
import pandas as pd
from ceruleo.transformation.features.operations import Divide, Sum
from ceruleo.transformation.features.selection import ByNameFeatureSelector
from ceruleo.transformation.functional.transformers import Transformer



class TestImputers():

    def test_Divide(self):

        divider = Divide()
        
        df1 = pd.DataFrame({
            'a': np.random.randn(5),
            'b': np.random.randn(5)
        })
        true_result = pd.DataFrame({
            'a': np.ones(5)*5,
            'b': np.ones(5)*3
        })
        df2 = df1.copy()
        df2['a'] = df1['a'] * 5 
        df2['b'] = df1['b'] * 3
        result_transformed = divider.transform([df2, df1])
        error = (true_result - result_transformed).sum().sum()
        assert(error < 0.0000000000001)

    def test_Sum(self):
        divider = Sum()
        
        df1 = pd.DataFrame({
            'a': np.ones(5)*3,
            'b': np.ones(5)*4
        })
        df2 = pd.DataFrame({
            'a': np.ones(5)*2,
            'b': np.ones(5)*3
        })
        true_result = pd.DataFrame({
            'a': np.ones(5)*5,
            'b': np.ones(5)*7
        })
        result_transformed = divider.transform([df2, df1])
        error = (result_transformed - true_result).sum().sum()
        assert(error < 0.0000000000001)

    def test_opeartorOverloading(self):
        pipe = ByNameFeatureSelector(features=['a', 'b'])
        pipe1 = pipe / 2 

        pipe2 = pipe / 3

        pipe = pipe1 + pipe2

        df = pd.DataFrame({
            'a': np.random.rand(5),
            'b': np.random.rand(5),
            'c': np.random.rand(5),
        })
        df1 = df*6

        df_result = (df1/2 + df1/3)[['a', 'b']]

        test_transformer = Transformer(pipelineX=pipe)
        test_transformer.fit([df1])
        result = test_transformer.transform(df1)[0]

        error = (df_result - result).sum().sum()
        assert(error < 0.00000001)

        




import numpy as np
from ceruleo.iterators.iterators import windowed_signal_generator
from ceruleo.utils.download import download


class TestWindowGeneratorTest:
    def test_signal_generator(self):
        N = 15
        X = np.zeros((N, 3))
        for j in range(3):
            X[:, j] = np.linspace(0, 25, num=N)

        y = np.linspace(0, 25, num=N)

        X_w, y_w = windowed_signal_generator(X, y, 3, 6, 1)

        assert y[3] == y_w
        assert X_w.shape[0] == 6
        assert X_w.shape[1] == 3
        assert X_w[-1, 0] == X[3, 0]

        X_w, y_w = windowed_signal_generator(X, y, 3, 6, 2)

        assert (np.squeeze(y[[3, 4]]) == np.squeeze(y_w)).all()

        X_w, y_w = windowed_signal_generator(X, y, 3, 6, 2, right_closed=False)

        assert X_w[-1, 0] == X[2, 0]
        assert (np.squeeze(y[[3, 4]]) == np.squeeze(y_w)).all()

def test_download():
    download("http://www.google.com", "test.html")



"""Visualization utilities for RUL estimation models


It is possible to visualize how it grows the unexploited lifetime grows as the conservative window size grows and how the unexpected breaks decrease as the conservative window size grows.


"""
import math
from typing import Dict, Iterable, List, Optional, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from ceruleo.dataset.transformed import TransformedDataset
from ceruleo.graphics.utils.curly_brace import curlyBrace
from ceruleo.results.results import (
    FittedLife,
    PredictionResult,
    models_cv_results,
    split_lives,
    unexpected_breaks,
    unexploited_lifetime,
)


def plot_lives(ds: TransformedDataset):
    """
    Plot each life

    Parameters:
        ds: A transformed dataset
    """
    fig, ax = plt.subplots()
    it = ds
    for _, y in it:
        ax.plot(y)
    return ax


def cv_plot_errors_wrt_RUL(bin_edges, error_histogram, **kwargs):
    """"""
    fig, ax = plt.subplots(**kwargs)
    labels = []
    heights = []
    xs = []
    yerr = []

    for i in range(len(error_histogram)):
        xs.append(i)
        heights.append(np.mean(error_histogram[i]))
        yerr.append(np.std(error_histogram[i]))
        labels.append(f"[{bin_edges[i]:.1f}, {bin_edges[i+1]:.1f})")

    ax.bar(height=heights, x=xs, yerr=yerr, tick_label=labels)
    ax.set_xlabel("RUL")
    ax.set_ylabel("RMSE")

    return ax


def _boxplot_errors_wrt_RUL_multiple_models(
    bin_edge: np.array,
    model_results: dict,
    ax=None,
    y_axis_label: Optional[str] = None,
    x_axis_label: Optional[str] = None,
    hold_out=False,
    **kwargs,
):
    def set_box_color(bp, color):
        plt.setp(bp["boxes"], color=color)
        plt.setp(bp["whiskers"], color=color)
        plt.setp(bp["caps"], color=color)
        plt.setp(bp["medians"], color=color)

    if ax is None:
        _, ax = plt.subplots(**kwargs)
    labels = []
    n_models = len(model_results)
    nbins = len(bin_edge) - 1

    for i in range(nbins):
        labels.append(f"[{bin_edge[i]:.1f}, {bin_edge[i+1]:.1f})")

    max_value = -np.inf
    min_value = np.inf
    colors = sns.color_palette("hls", n_models)
    for model_number, model_name in enumerate(model_results.keys()):
        model_data = model_results[model_name]
        hold_out = model_data.n_folds == 1
        if hold_out:
            for errors in model_data.errors:
                min_value = min(min_value, np.min(errors))
                max_value = max(max_value, np.max(errors))
        else:
            min_value = min(min_value, np.min(model_data.mean_error))
            max_value = max(max_value, np.max(model_data.mean_error))
        positions = []
        for i in range(nbins):
            positions.append((model_number * 0.5) + (i * n_models))
        if hold_out:
            box = ax.boxplot(
                np.array(model_data.errors, dtype=object),
                positions=positions,
                widths=0.2,
            )
        else:
            box = ax.boxplot(model_data.mean_error, positions=positions, widths=0.2)
        set_box_color(box, colors[model_number])
        ax.plot([], c=colors[model_number], label=model_name)

    ticks = []
    for i in range(nbins):
        x = np.mean(
            [(model_number * 0.5) + (i * n_models) for model_number in range(n_models)]
        )
        ticks.append(x)

    max_x = np.max(ticks) + 1
    ax.set_xlabel("RUL" + ("" if x_axis_label is None else x_axis_label))
    ax.set_ylabel(r"$y - \hat{y}$" + ("" if y_axis_label is None else y_axis_label))
    ax.set_xticks(ticks)
    ax.set_xticklabels(labels)
    ax.legend()
    ax2 = ax.twinx()
    ax2.set_xlim(ax.get_xlim())
    ax2.set_ylim(ax.get_ylim())
    curlyBrace(
        ax.figure, ax2, (max_x, 0), (max_x, min_value), str_text="Over estim.", c="#000"
    )
    ax2.get_xaxis().set_visible(False)
    ax2.get_yaxis().set_visible(False)
    curlyBrace(
        ax.figure,
        ax2,
        (max_x, max_value),
        (max_x, 0),
        str_text="Under estim.",
        c="#000",
    )

    return ax


def boxplot_errors_wrt_RUL(
    results_dict: Dict[str, List[PredictionResult]],
    nbins: int,
    y_axis_label: Optional[str] = None,
    x_axis_label: Optional[str] = None,
    ax=None,
    **kwargs,
) -> matplotlib.axes.Axes:
    """
    Boxplots of difference between true and predicted RUL over Cross-validated results


    Parameters:
        results_dict: Dictionary with the results of the fitted models
        nbins: Number of bins to divide the
        y_axis_label: Optional string to be added to the y axis
        x_axis_label: Optional string to be added to the x axis
        ax: Optional axis in which the plot will be drawed.
            If an axis is not provided, it will create one.

    Keyword arguments:
        **kwargs

    Return:
        The axis in which the plot has been made
    """
    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    else:
        fig = ax.figure

    bin_edges, model_results = models_cv_results(results_dict, nbins)
    return _boxplot_errors_wrt_RUL_multiple_models(
        bin_edges,
        model_results,
        fig=fig,
        ax=ax,
        y_axis_label=y_axis_label,
        x_axis_label=x_axis_label,
    )


def _cv_barplot_errors_wrt_RUL_multiple_models(
    bin_edges,
    model_results: dict,
    ax=None,
    y_axis_label: Optional[str] = None,
    x_axis_label: Optional[str] = None,
    color_palette: str = "hls",
    bar_width: float=1/1.5,
    **kwargs,
) -> Tuple[matplotlib.figure.Figure ,matplotlib.axes.Axes]:
    """
    Plot the barplots given the errors

    Parameters:
        bin_edges: np.ndarray:
        model_results: Dictionary with the results
        ax: Axis
        y_axis_label: Y Label
        x_axis_label: X Label

    Returns:
        The plot axis
    """
    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    fig  = ax.figure
    labels = []
    n_models = len(model_results)
    nbins = len(bin_edges) - 1

   

    for i in range(nbins):
        labels.append(f"[{bin_edges[i]:.1f}, {bin_edges[i+1]:.1f})")

    colors = sns.color_palette(color_palette, n_models)
    
    mean_data = []
    std_data = []
    model_names = []
    for model_number, model_name in enumerate(model_results.keys()):
        model_data = model_results[model_name]
        mean_data.append(np.mean(model_data.mae, axis=0))
        std_data.append(np.std(model_data.mae, axis=0))
        model_names.append(model_name)

    model_names = np.array(model_names)

    bar_group_width = n_models*(bar_width+1)
    group_separation = int(bar_group_width/2)
    mean_data = np.vstack(mean_data)
    std_data = np.vstack(std_data)
    n_models, n_bins = mean_data.shape
    indices = np.argsort(mean_data[:,  0])
    for i in range(n_bins):
        mean_data[:, i] = mean_data[indices, i]
        std_data[:, i] = std_data[indices, i]
    
    for model_name, model_index in zip(model_names[indices], range(n_models)):    
        
        positions =  model_index+np.array(range(n_bins))  * (bar_group_width + group_separation)

        rect = ax.bar(
            positions,
            mean_data[model_index, :],
            yerr=std_data[model_index, :],
            label=model_name,
            width=bar_width,
            color=colors[model_index],
        )

    ticks = []
    dx = 0
    for i in range(nbins):
        ticks.append( dx + bar_group_width/2)
        dx += bar_group_width + group_separation

    ax.set_xlabel("RUL" + ("" if x_axis_label is None else x_axis_label))
    ax.set_ylabel(r"$y - \hat{y}$" + ("" if y_axis_label is None else y_axis_label))
    ax.set_xticks(ticks)
    ax.set_xticklabels(labels)
    ax.legend()

    return ax


def barplot_errors_wrt_RUL(
    results_dict: Dict[str, List[PredictionResult]],
    nbins: int,
    y_axis_label=None,
    x_axis_label=None,
    ax=None,
    color_palette: str = "hls",
    **kwargs,
):
    """
    Barlots of difference between true and predicted RUL

    Parameters:
        results_dict: Dictionary with the results for each model
        nbins: Number of bins in wich divide the RUL target
        y_axis_label: Y label
        x_axis_label: X label
        ax: Axis
        color_palette: 

    """
    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    bin_edges, model_results = models_cv_results(results_dict, nbins)
    return _cv_barplot_errors_wrt_RUL_multiple_models(
        bin_edges,
        model_results,
        ax=ax,
        y_axis_label=y_axis_label,
        x_axis_label=x_axis_label,
        color_palette=color_palette,
    )


def _cv_shadedline_plot_errors_wrt_RUL_multiple_models(
    bin_edges: np.array,
    model_results: dict,
    ax: Optional[matplotlib.axes.Axes]=None,
    y_axis_label: Optional[str] =None,
    x_axis_label: Optional[str] =None,
    **kwargs,
):
    """Plot a shaded regions for each model

    Parameters:
        bin_edges: Bin Edges
        model_results: Dictionary with the results for the model
        y_axis_label: Y label
        x_axis_label: X label
    """
    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    labels = []
    n_models = len(model_results)
    nbins = len(bin_edges) - 1

    width = 1.0 / n_models

    for i in range(nbins):
        labels.append(f"[{bin_edges[i]:.1f}, {bin_edges[i+1]:.1f})")
    max_value = -np.inf
    min_value = np.inf
    colors = sns.color_palette("hls", n_models)
    for model_number, model_name in enumerate(model_results.keys()):
        model_data = model_results[model_name]
        hold_out = model_data.n_folds == 1
        if hold_out:
            for errors in model_data.errors:
                min_value = min(min_value, np.min(errors))
                max_value = max(max_value, np.max(errors))
        else:
            min_value = min(min_value, np.min(model_data.mean_error))
            max_value = max(max_value, np.max(model_data.mean_error))

        positions = []
        for i in range(nbins):
            positions.append((model_number * width) + (i * n_models))

        if not hold_out:
            mean_error = np.mean(model_data.mean_error, axis=0)
            std_error = np.std(model_data.mean_error, axis=0)
        else:
            mean_error = np.array([np.mean(e, axis=0) for e in model_data.errors])
            std_error = np.array([np.std(e, axis=0) for e in model_data.errors])

        rect = ax.plot(
            positions, mean_error, label=model_name, color=colors[model_number]
        )
        ax.fill_between(
            positions,
            mean_error - std_error,
            mean_error + std_error,
            alpha=0.3,
            color=colors[model_number],
        )

    ticks = []
    for i in range(nbins):
        x = np.mean(
            [(model_number * 0.5) + (i * n_models) for model_number in range(n_models)]
        )
        ticks.append(x)

    ax.set_xlabel("RUL" + ("" if x_axis_label is None else x_axis_label))
    ax.set_ylabel(r"$y - \hat{y}$" + ("" if y_axis_label is None else y_axis_label))
    ax.set_xticks(ticks)
    ax.set_xticklabels(labels)
    ax.legend()
    max_x = np.max(ticks) + 1
    ax2 = ax.twinx()
    ax2.set_xlim(ax.get_xlim())
    ax2.set_ylim(ax.get_ylim())
    curlyBrace(
        ax2.figure, ax2, (max_x, 0), (max_x, min_value), str_text="Over estim.", c="#000"
    )
    ax2.get_xaxis().set_visible(False)
    ax2.get_yaxis().set_visible(False)
    curlyBrace(
        ax2.figure, ax2, (max_x, max_value), (max_x, 0), str_text="Under estim.", c="#000"
    )

    return ax


def shadedline_plot_errors_wrt_RUL(
    results_dict: dict,
    nbins: int,
    y_axis_label: Optional[str] =None,
    x_axis_label: Optional[str] =None,
    ax: Optional[matplotlib.axes.Axes] =None,
    **kwargs,
) -> matplotlib.axes.Axes:
    """
    Shaded line

    Parameters:
        results_dict: Dictionary with the results for the model
        nbins: Number of bins
        y_axis_label: Y label
        x_axis_label: X label
        ax: Plot axis 

    Returns:
        The plot axis
    """

    if ax is None:
        fig, ax = plt.subplots(**kwargs)

    bin_edges, model_results = models_cv_results(results_dict, nbins)
    return _cv_shadedline_plot_errors_wrt_RUL_multiple_models(
        bin_edges,
        model_results,
        ax=ax,
        bins=nbins,
        y_axis_label=y_axis_label,
        x_axis_label=x_axis_label,
    )


def plot_unexploited_lifetime(
    results_dict: dict,
    max_window: int,
    n: int,
    ax=None,
    units: Optional[str] = "",
    add_shade: bool = True,
    **kwargs,
):
    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    n_models = len(results_dict)
    colors = sns.color_palette("hls", n_models)
    for i, model_name in enumerate(results_dict.keys()):
        m, ulft, std_ul = unexploited_lifetime(results_dict[model_name], max_window, n)
        ax.plot(m, ulft, label=model_name, color=colors[i])
        if add_shade:
            ax.fill_between(m, ulft+std_ul, ulft-std_ul, alpha=0.1, color=colors[i])
    ax.legend()
    ax.set_title("Unexploited lifetime")
    ax.set_xlabel("Fault window size" + units)
    ax.set_ylabel(units)
    return ax


def plot_unexpected_breaks(
    results_dict: dict,
    max_window: int,
    n: int,
    ax: Optional[matplotlib.axes.Axes] = None,
    units: Optional[str] = "",
    add_shade: bool = True,
    **kwargs,
) -> matplotlib.axes.Axes:
    """
    Plot the risk of unexpected breaks with respect to the maintenance window

    Parameters:
        results_dict: Dictionary with the results
        max_window: Maximum size of the maintenance windows
        n: Number of points used to evaluate the window size
        ax: axis on which to draw, by default None
        units: Units to use in the xlabel, by default ""

    Returns:
        The axis in which the plot was made
    """
    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    n_models = len(results_dict)
    colors = sns.color_palette("hls", n_models)
    for i, model_name in enumerate(results_dict.keys()):
        m, mean_ub, std_ub = unexpected_breaks(results_dict[model_name], max_window, n)
        ax.plot(m, mean_ub, label=model_name, color=colors[i])
        if add_shade:
            ax.fill_between(m, mean_ub+std_ub, mean_ub-std_ub, alpha=0.1, color=colors[i])

    ax.set_title("Unexpected breaks")
    ax.set_xlabel("Fault window size" + units)
    ax.set_ylabel("Risk of breakage")
    ax.legend()
    return ax




def plot_J_Cost(
    results: Dict[str, List[PredictionResult]],
    window: int,
    step: int,
    ax=None,
    ratio_min: float = 1 / 120,
    ratio_max: float = 1 / 5,
    ratio_n_points: int = 50,
):
    def label_formatter(x):
        UB_c = 1 / x
        UL_c = 1
        if x == 0:
            return ""
        return f"{int(UB_c)}:{int(UL_c)}"

    if ax is None:
        fig, ax = plt.subplots(figsize=(17, 5))

    ratio = np.linspace(ratio_min, ratio_max, ratio_n_points)
    n_models = len(results)
    colors = sns.color_palette("hls", n_models)
    for i, model_name in enumerate(results.keys()):
        window_ub, mean_ub, std_ub = unexpected_breaks(results[model_name], window_size=window, step=step)
        window_ul, mean_ul, std_ul = unexploited_lifetime(results[model_name], window_size=window, step=step)

        v = []

        labels = []
        from uncertainties import unumpy
   
        for r in ratio:
            UB_c = 1.0
            UL_c = UB_c * r
            
            v.append(np.mean(unumpy.uarray(mean_ub, std_ub) * UB_c + unumpy.uarray(mean_ul, std_ul) * UL_c))

            labels.append(f"{int(UB_c)}:{UL_c}")

      
  
        mean = unumpy.nominal_values(v)
        std = unumpy.std_devs(v)
        ax.plot(ratio, mean, "-o", label=model_name, color=colors[i])
        ax.fill_between(ratio, mean+std, mean-std, color=colors[i], alpha=0.2)

    ticks = ax.get_xticks().tolist()
    ticks.append(ratio[0])
    ax.set_xticks(ticks)
    ax.set_xticklabels([label_formatter(x) for x in ax.get_xticks()])
    ax.set_xlabel(
        "Ratio between UL and UB. How many minutes of UL are equal to 1 breakage"
    )
    ax.set_ylabel("J")
    return ax


def plot_life(
    life: FittedLife,
    ax: Optional[matplotlib.axes.Axes]=None,
    units: Optional[str] = "",
    markersize: float = 0.7,
    add_fitted: bool = False,
    plot_target: bool = True,
    add_regressed: bool = True,
    start_x:int= 0,
    label:str = '',
    **kwargs,
) -> matplotlib.axes.Axes:
    """
    Plot a single life

    Parameters:
        life: A fitted life
        ax: The axis where to plot
        units : Optional[str], optional
            _description_, by default ""
        markersize: Size of the marker
        add_fitted: Wether to add the LS fitted line to the points
        plot_target: Wether to plot the true RUL values
        add_regressed:
        start_x: Initial point of the time-indepedent feature to plot
        label: 

    Returns:
        The plot axis
    """
    if ax is None:
        _, ax = plt.subplots(1, 1, **kwargs)

    time = life.time

    ax.plot(
        life.time[start_x: len(life.y_pred)],
        life.y_pred[start_x:],
        "o",
        label=f"Predicted {label}",
        markersize=markersize,
    )
    if plot_target:
        ax.plot(life.time, life.y_true, label="True", linewidth=3)
    if add_regressed and life.y_true[-1] > 0:
            time1 = np.hstack((time[-1], time[-1] + life.y_true[-1]))
            ax.plot(time1, [life.y_true[-1], 0], label="Regressed true")
    if add_fitted:
        #time1 = np.hstack(
        #    (time[len(life.y_pred) - 1], time[len(life.y_pred) - 1] + life.y_pred[-1])
        #)
        #ax.plot(time1, [life.y_pred[-1], 0], label="Projected end")
        ax.plot(
           life.time,
           life.y_pred_fitted,
           label="Picewise fitted",
        )
        ax.plot(
           life.time,
           life.y_true_fitted,
           label="Picewise fitted",
        )

    ax.set_ylabel(units)
    ax.set_xlabel(units)
    _, max = ax.get_ylim()
    ax.set_ylim(0 - max / 10, max)
    legend = ax.legend(markerscale=15,)



    return ax


def plot_predictions_grid(
    results: Union[PredictionResult, List[PredictionResult]],
    ncols: int = 3,
    alpha: float =1.0,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    **kwargs,
):
    """
    Plot a matrix of predictions

    Parameters:
        results: Dictionary with the results
        ncols: Number of colmns in the plot
        alpha: Opacity of the predicted curves
        xlabel: Xlabel
        ylabel: YLabel

    Return:
        The axis on which the plot has been made
    """

    def linear_to_subindices(i, ncols):
        row = int(i / ncols)
        col = i % ncols
        return row, col

    if isinstance(results, PredictionResult):
        results = [results]

    init = False

    for model_results in results:
        lives_model = split_lives(model_results)
        NROW = math.ceil(len(lives_model) / ncols)
        if not init:
            fig, ax = plt.subplots(NROW, ncols, squeeze=False, **kwargs)

        for i, life in enumerate(lives_model):
            row, col = linear_to_subindices(i, ncols)

            if not init:
                ax[row, col].plot(life.time, life.y_true, label="True")

            ax[row, col].plot(
                life.time, life.y_pred, label=model_results.name, alpha=alpha
            )
            if xlabel is not None:
                ax[row, col].set_xlabel(xlabel)
            if ylabel is not None:
                ax[row, col].set_ylabel(ylabel)
        init = True
    for j in range(len(lives_model), NROW * ncols):
        row, col = linear_to_subindices(j, ncols)
        fig.delaxes(ax[row, col])

    for a in ax.flatten():
        a.legend()
    return ax


def plot_predictions(
    result: Union[PredictionResult, Tuple[np.ndarray, np.ndarray]],
    *,
    ax:Optional[matplotlib.axes.Axes]=None,
    units: str = "Hours [h]",
    markersize: float = 0.7,
    marker: str = 'o',
    plot_fitted: bool  = True,
    model_name:str = '',
    **kwargs,
) -> matplotlib.axes.Axes:
    """
    Plots the predicted and the true remaining useful lives

    Parameters:
        result: A PredictionResult object or a tuple with (y_true, y_predicted)
        ax:  Axis to plot. If it is missing a new figure will be created
        units: Units of time to be used in the axis labels
        marker: Marker type
        markersize: The size of the marker
        plot_fitted: Wether to plot a LS line
        model_name: Name of the model
        

    Returns:
        The axis on which the plot has been made
    """
    if ax is None:
        _, ax = plt.subplots(1, 1, **kwargs)


    if isinstance(result, PredictionResult):
        y_predicted = result.predicted_RUL
        y_true = result.true_RUL
    else:
        y_true, y_predicted = result
    ax.plot(y_predicted, marker, label=f"Predicted {model_name}", markersize=markersize)
    ax.plot(y_true, label="True")
    x = 0


    if plot_fitted:
        try:
            fitted = np.hstack([life.y_pred_fitted for life in split_lives(result)])
            ax.plot(fitted, label='Fitted')
            
        except:
            
            pass
    ax.set_ylabel(units)
    ax.set_xlabel(units)
    legend = ax.legend()
    for l in legend.legend_handles:
        l.set_markersize(6)


    return ax



import logging
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from ceruleo.transformation import TransformerStep

logger = logging.getLogger(__name__)





class ByNameFeatureSelector(TransformerStep):
    """Select a subset of feature by name

    Parameters:
            features: Feature name or List of features name to select
    """
    def __init__(self, *, features:Union[str, List[str]]= [], name: Optional[str] = None):
        super().__init__(name=name)
        if isinstance(features, str):
            features = [features]
        self.features = features
        self.features_indices = None
        self.features_computed_ = []

    def partial_fit(self, df, y=None):
        if len(self.features) > 0:
            features = [f for f in self.features if f in set(df.columns)]
        else:
            features = list(set(df.columns))

        if len(self.features_computed_) == 0:
            self.features_computed_ = features
        else:
            self.features_computed_ = [
                f for f in self.features_computed_ if f in features
            ]
        return self

    def fit(self, df:pd.DataFrame, y=None):
        """ 
        Find the indices of the features to select

        Parameters:
            df: DataFrame containing the input life
        """
        if len(self.features) > 0:
            features = [f for f in self.features if f in set(df.columns)]
        else:
            features = list(set(df.columns))
        self.features_computed_ = sorted(features)
        return self
        return X.loc[:, self.features_computed_].copy()

    def transform(self, X:pd.DataFrame) -> pd.DataFrame:
        """ 
        Transform the input life

        Parameters:
            X: The input life to be transformed

        Returns:
            A new DataFrame containing only the selected features
        """
        return X.loc[:, self.features_computed_].copy()

    @property
    def n_features(self):
        return len(self.features_computed_)

    def description(self):
        name = super().description()
        return (name, self.features_computed_)

    def __str__(self):
        name, f = self.description()
        features = ', '.join(f)[:10]
        return f'{name} : [{features}]'
        


class PositionFeatures(TransformerStep):
    """
    Reorder the features of the input life

    Parameters:
        features: Dictionary containing the features to reorder and their new position
        name: Name of the step, by default None
    """
    def __init__(self, *, features: dict, name: Optional[str] = None):
        super().__init__(name=name)
        self.features = features

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Transform the input life by reordering the features

        Parameters:
            X: The input life to be transformed
        
        Returns:
            A new DataFrame containing the features in the order specified in the constructor
        """
        cols = list(X.columns)
        for name, pos in self.features.items():
            a, b = cols.index(name), pos
            cols[b], cols[a] = cols[a], cols[b]
            X = X[cols]
        return X


class DiscardByNameFeatureSelector(TransformerStep):
    """
    Remove a list of features from the input life

    Parameters:
        features: List of features to discard
        name: Name of the step, by default None
    """
    def __init__(self, *, features: List=[], name: Optional[str] = None):
        super().__init__(name=name)
        self.features = features
        self.features_indices = None

    def fit(self, df:pd.DataFrame, y=None):
        """
        Find the indices of the features to discard

        Parameters:
            df: DataFrame containing the set of features to discard
        """
        self.feature_columns = [f for f in df.columns if f not in self.features]
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Transform the input life

        Parameters:
            X: The input life to be transformed
        
        Returns:
            A new DataFrame containing only the features not in the list of features to discard
        """
        return X.loc[:, self.feature_columns]

    @property
    def n_features(self):
        return len(self.features)


class PandasVarianceThreshold(TransformerStep):
    """ 
    Remove features with variance lower than a variance threshold inserted in input

    Parameters:
        min_variance: Minimum variance threshold
        name: Name of the step, by default None
    """
    def __init__(self, *, min_variance: float, name: Optional[str] = None):
        super().__init__(name=name)
        self.min_variance = min_variance
        self.selected_columns_ = None

    def partial_fit(self, X:pd.DataFrame, y=None):
        """ 
        Find the indexes of the features with variance higher than the threshold

        Parameters:
            X: DataFrame containing the input life
        """
        variances_ = X.var(skipna=True)
        partial_selected_columns_ = X.columns[variances_ > self.min_variance]
        if (
            self.selected_columns_ is not None
            and len(partial_selected_columns_) < len(self.selected_columns_) * 0.5
        ):
            logger.warning(type(self).__name__)
            logger.warning(
                f"Life removed more than a half of the columns. Shape {X.shape}"
            )
            logger.warning(
                f"Current: {len(self.selected_columns_)}. New ones: {len(partial_selected_columns_)}"
            )
        if self.selected_columns_ is None:
            self.selected_columns_ = partial_selected_columns_
        else:
            self.selected_columns_ = [
                f for f in self.selected_columns_ if f in partial_selected_columns_
            ]
        if len(self.selected_columns_) == 0:
            logger.warning(type(self).__name__)
            logger.warning("All features were removed")
        return self

    def fit(self, X:pd.DataFrame, y=None):
        """ 
        Find the indexes of the features with variance higher than the threshold

        Parameters:
            X: DataFrame containing the input life
        """
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        self.variances_ = X.var(skipna=True)
        self.selected_columns_ = X.columns[self.variances_ > self.min_variance]
        logger.debug(
            f"Dropped columns {[c for c in X.columns if c not in self.selected_columns_]}"
        )
        return self

    def transform(self, X:pd.DataFrame, y=None) -> pd.DataFrame:
        """ 
        Transform the input life

        Parameters:
            X: The input life to be transformed

        Returns:
            A new life containing only the features with variance higher than the threshold
        """
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        return X[self.selected_columns_].copy()


class NullProportionSelector(TransformerStep):
    """ 
    Remove features with null proportion higher than a threshold inserted in input

    Parameters:
        max_null_proportion: Maximum null proportion threshold
        name: Name of the step, by default None
    """
    def __init__(self, *, max_null_proportion: float, name: Optional[str] = None):
        super().__init__(name=name)
        self.max_null_proportion = max_null_proportion
        self.selected_columns_ = None

    def partial_fit(self, X:pd.DataFrame, y=None):
        """ 
        Find the indexes of the features with null proportion lower than the threshold

        Parameters:
            X: DataFrame containing the input life
        """
        null_proportion = X.isnull().mean()

        partial_selected_columns_ = X.columns[
            null_proportion < self.max_null_proportion
        ]
        if (
            self.selected_columns_ is not None
            and len(partial_selected_columns_) < len(self.selected_columns_) * 0.5
        ):
            logger.warning(type(self).__name__)

        if self.selected_columns_ is None:
            self.selected_columns_ = partial_selected_columns_
        else:
            self.selected_columns_ = [
                f for f in self.selected_columns_ if f in partial_selected_columns_
            ]
        if len(self.selected_columns_) == 0:
            logger.warning(type(self).__name__)
            logger.warning("All features were removed")
        return self

    def fit(self, X:pd.DataFrame, y=None):
        """
        Find the indexes of the features with null proportion lower than the threshold

        Parameters:
            X: DataFrame containing the input life
        """
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        self.null_proportion = X.isnull().mean()
        self.selected_columns_ = X.columns[
            self.null_proportion < self.max_null_proportion
        ]
        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Transform the input life

        Parameters:
            X: The input life to be transformed
        
        Returns:
            A new life containing only the features with null proportion lower than the threshold
        """
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        return X[self.selected_columns_].copy()


class  MatchFeatureSelector(TransformerStep):
    """Select all the features that match a pattern

    Parameters:
        pattern: Pattern to match
    """
    def __init__(self, *, pattern:str, name: Optional[str] = None):
        super().__init__(name=name)
        self.pattern = pattern
        self.selected_columns_ = None


    def partial_fit(self, df: pd.DataFrame, y=None):

        """ 
        Find the features matching the pattern

        Parameters:
            df: DataFrame containing the entire set of features 
        """

        if self.selected_columns_ is None:
            self.selected_columns_ = [f for f in df.columns if self.pattern in f ]

        return self

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Transform the input life

        Parameters:
            X: The input life to be transformed

        Returns:
            A new life with the same index as the input with the missing values replaced by the value in the succesive timestamp 
        """
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input array must be a data frame")
        return X[self.selected_columns_].copy()


class ByTypeFeatureSelector(TransformerStep):
    """Select a subset of feature by type

    Parameters:
            type_: Data type to be selected, by default []
    """
    def __init__(self, *, type_:Union[str, List]= [], name: Optional[str] = None):
        super().__init__(name=name)

        self.features = []
        self.type = type_

    def partial_fit(self, df, y=None):
        if len(self.features) == 0:            
            self.features = df.select_dtypes(include=self.type).columns      
        return self

    def fit(self, df, y=None):
        if len(self.features) == 0:
            self.features = df.select_dtypes(include=self.type).columns

        return self

    def transform(self, X:pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input life

        Parameters:
            X: The input life to be transformed

        Returns:
            A new DataFrame containing only the features of the selected type
        """
        return X.loc[:, self.features].copy()

    @property
    def n_features(self):
        return len(self.features)

    def description(self):
        name = super().description()
        return (name, self.features)

    def __str__(self):
        name, f = self.description()
        features = ', '.join(f)[:10]
        return f'{name} : [{features}]'


import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from ceruleo.transformation import TransformerStep


class Subsample(TransformerStep):
    def __init__(self, *, steps: int, **kwargs):
        super().__init__(**kwargs)
        self.steps = steps

    def transform(self, X: pd.DataFrame):
        return X.iloc[:: self.steps, :]


class IndexMeanResampler(TransformerStep):
    """Resample 
    When the index of the run-to-failure cycle is a time feature

    Parameters:
        rule: Time frequency or rule according to which the data should be resampled
    """
    def __init__(self, *, rule: str,  **kwargs):
        super().__init__(**kwargs)
        self.rule = rule
    
    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Apply the transformation to the input life

        Parameters:
            X: The input life
        
        Returns: 
            The resampled DataFrame
        """
        return X.resample(self.rule).mean().dropna()
        


class SubsamplerTransformer(TransformerStep):
    """IntegerIndexResamplerTransformer

    Resample the time series with an integer index and interpolate linearly the values

    Parameters:
        time_feature: Time feature
        steps:  Number of steps
        drop_time_feature: Drop the time feature
    """

    def __init__(self, *args, time_feature: str, steps: int, drop_time_feature: bool):
        super().__init__(*args)
        self._time_feature_name = time_feature
        self._time_feature = None
        self.steps = steps
        self.drop_time_feature = drop_time_feature

    def partial_fit(self, X: pd.DataFrame):
        """Obtain the name of the feature used as time

        Parameters:
            X: The current time-series to be fitted

        Returns:
            Instance of class IntegerIndexResamplerTransformer

        """
        if self._time_feature is None:
            self._time_feature = self.find_feature(X, self._time_feature_name)
            if self._time_feature is None:
                raise ValueError("Time feature not found")

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Apply the transformation to the input life

        Parameters:
            X: The input life
        
        Returns: 
            The resampled DataFrame
        """
        X = X.groupby(X[self._time_feature] // self.steps, sort=False).mean()
        if self.drop_time_feature:
            X = X.drop(columns=[self._time_feature])

        return X


class IntegerIndexResamplerTransformer(TransformerStep):
    """
    Resample the time series with an integer index and interpolate linearly the values

    Parameters:
        time_feature: Time feature
        steps: Number of steps
        drop_time_feature: Drop the time feature
    """

    def __init__(self, *args, time_feature: str, steps: int, drop_time_feature: bool):
        super().__init__(*args)
        self._time_feature_name = time_feature
        self._time_feature = None
        self.steps = steps
        self.drop_time_feature = drop_time_feature

    def partial_fit(self, X: pd.DataFrame):
        """
        Obtain the name of the feature used as time

        Parameters:
            X: The current time-series to be fitted

        Returns:
            Instance of class IntegerIndexResamplerTransformer
        """
        if self._time_feature is None:
            self._time_feature = self.find_feature(X, self._time_feature_name)
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Apply the transformation to the input life

        Parameters:
            X: The input life
        
        Returns: 
            The resampled DataFrame
        """
        Told = X[self._time_feature].values
        Tnew = np.arange(Told.min(), Told.max(), self.steps)
        new_columns = X.columns.values
        if self.drop_time_feature:
            new_columns = np.delete(
                new_columns, np.where(new_columns == self._time_feature)
            )

        new_X = pd.DataFrame(index=Tnew, columns=new_columns)
        for c in new_columns:
            try:
                F = interp1d(Told, X[c])
                new_X[c] = F(Tnew)
            except ValueError:
                # https://stackoverflow.com/questions/62015823/interpolating-categorical-data-in-python-nearest-previous-value/62016097#62016097
                y = X[c].values
                f = interp1d(
                    Told,
                    range(len(y)),
                    kind="nearest",
                    fill_value=(0, len(y) - 1),
                    bounds_error=False,
                )
                y_idx = f(Tnew)
                new_X[c] = [y[int(i)] for i in y_idx]
        return new_X



"""This module provides interoperability 

Scikit learned models can be used wit the ceruleo Transformers


The `TimeSeriesWindowTransformer`  is a scikit-learn transformers that takes
a transformer and iterator paramaters, to build a `WindowedDatasetIterator`
and generate the X and y.

Since the X and Y are generated by one scikit-learn transformer step, `EstimatorWrapper`
takes the (X,y) output of a `TimeSeriesWindowTransformer` and calls the rest of the scikit-learn
pipeline spreading the (X,y)  to positional parameters.

Finally `CeruleoRegressor`, is a class similar to the `sklearn.compose.TransformedTargetRegressor`.
Takes the transformer, and a regressor scikit-learn pipeline or model, and automatically builds
a scikit-learn pipeline using `WindowedDatasetIterator` and `EstimatorWrapper`.
"""
import logging
from typing import Optional, Tuple

import numpy as np
import pandas as pd
from ceruleo.dataset.ts_dataset import AbstractPDMDataset
from ceruleo.iterators.batcher import Batcher
from ceruleo.iterators.iterators import (
    NotWeighted,
    SampleWeight,
    WindowedDatasetIterator,
)
from ceruleo.iterators.shufflers import AbstractShuffler, NotShuffled
from ceruleo.transformation.functional.transformers import Transformer

import sklearn.pipeline as sk_pipeline
from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.metrics import mean_squared_error
from sklearn.metrics._scorer import get_scorer
from typing import List

logger = logging.getLogger(__name__)


class EstimatorWrapper(TransformerMixin, BaseEstimator):
    """
    Wrapper around sklearn estimators to allow calling the fit and predict

    The transformer keeps the X and y together. This wrapper
    divide the X,y and call the fit(X,y) and predict(X,y) of the estimator

    Parameters:
        estimator: A scikit-learn estimator
    """

    def __init__(self, estimator: BaseEstimator):
        self.estimator = estimator

    def fit(self, Xy: Tuple[np.ndarray, np.ndarray], y=None, **fit_params):
        X, y = Xy
        self.estimator.fit(X, y, **fit_params)
        return self

    def predict(self, Xy, **transform_params):
        X, y = Xy
        return self.estimator.predict(X, **transform_params)


class TimeSeriesWindowTransformer(TransformerMixin, BaseEstimator):
    """
    A scikit-learn transformer for obtaining a windowed time-series from the run-to-cycle failures

    Parameters:
        transformer: A scikit-learn transformer
        window_size: Window size of the iterator
        step: Stride of the iterators
        horizon: Horizon of the predictions
        shuffler: Shuffler of the data
        sample_weight: Sample weights callable
        right_closed: Wether to include the last point in each sliding window
        padding: Wether to pad when the lookback window is not complete
    """

    def __init__(
        self,
        transformer: Transformer,
        window_size: int,
        step: int = 1,
        horizon: int = 1,
        shuffler: AbstractShuffler = NotShuffled(),
        sample_weight: SampleWeight = NotWeighted(),
        right_closed: bool = True,
        padding: bool = True,
    ):
        self.transformer = transformer
        self.window_size = window_size
        self.horizon = horizon
        self.step = step
        self.shuffler = shuffler
        self.sample_weight = sample_weight
        self.right_closed = right_closed
        self.padding = padding

    def fit(self, dataset: AbstractPDMDataset):
        """
        Fit the transformer with the given dataset

        Parameters:
            dataset: Dataset to fit the transformer
        """
        self.transformer.fit(dataset)
        return self

    def _iterator(self, dataset: AbstractPDMDataset):
        return WindowedDatasetIterator(
            dataset.map(self.transformer),
            self.window_size,
            self.step,
            self.horizon,
            shuffler=self.shuffler,
            sample_weight=self.sample_weight,
            right_closed=self.right_closed,
            padding=self.padding,
        )

    def transform(self, dataset: AbstractPDMDataset):
        X, y, sw = self._iterator(dataset).get_data()
        return X, y.ravel()

    def true_values(self, dataset: AbstractPDMDataset):
        X, y, sw = self._iterator(dataset).get_data()
        return y.ravel()

    def get_params(self, deep=None):
        params = super().get_params(deep)
        if deep:
            params["ts_window_transformer__transformer"] = self.transformer.get_params(
                deep
            )
        return params


class CeruleoRegressor(RegressorMixin, BaseEstimator):
    """
    A regressor wrapper similar to sklearn.compose.TransformedTargetRegressor


    Parameters:
        features_transformer: The transformer
        regressor: A scikit-learn regressor
    """

    def __init__(
        self, ts_window_transformer: TimeSeriesWindowTransformer, regressor, **kwargs
    ):
        self.ts_window_transformer = ts_window_transformer
        self.regressor = regressor
        self._update()

    def _update(self):
        self.wrapped_regressor = EstimatorWrapper(self.regressor)

        self.pipe = sk_pipeline.make_pipeline(
            self.ts_window_transformer, self.wrapped_regressor
        )

    def fit(self, dataset: AbstractPDMDataset, y=None):
        self.pipe.fit(dataset)
        return self

    def predict(self, data):
        if isinstance(data, np.ndarray):
            return self.regressor.predict(data)
        else:
            return self.pipe.predict(data)

    def score(self, dataset, y=None):
        X, y = self.ts_window_transformer.transform(dataset)
        return super().score(dataset, y)

    def set_params(self, **kwargs):
        super().set_params(**kwargs)
        self._update()
        return self


class CeruleoMetricWrapper:
    """
    A wrapper around sklearn metrics

        Example:
        '''
            grid_search = GridSearchCV(
                estimator=regressor_gs,
                param_grid={
                    'regressor': [RandomForestRegressor(max_depth=5)]
                },
                scoring=CeruleoMetricWrapper('neg_mean_absolute_error')
            )
        '''
    """

    def __init__(self, scoring):
        if callable(scoring):
            self.scorers = scoring
        elif isinstance(scoring, str):
            self.scorers = get_scorer(scoring)

    def __call__(self, estimator: CeruleoRegressor, dataset, y=None):
        X, y = estimator.ts_window_transformer.transform(dataset)
        return self.scorers(estimator, X, y)


def train_model(
    model,
    train_iterator: WindowedDatasetIterator,
    val_windowed_iterator: Optional[WindowedDatasetIterator] = None,
    **fit_kwargs,
):
    """
    Fit the model with the given dataset iterator
    Parameters:
        train_iterator: Training Iterator

    Keyword arguments:
        fit_kwargs: Arguments for the fit method

    Returns:
        A SKLearn model
    """
    X, y, sample_weight = train_iterator.get_data()

    params = {}
    try:
        from xgboost import XGBRegressor

        if val_windowed_iterator is not None and isinstance(model, XGBRegressor):
            X_val, y_val, _ = val_windowed_iterator.get_data()
            fit_kwargs.update({"eval_set": [(X_val, y_val)]})
    except Exception as e:
        logger.error(e)

    return model.fit(X, y.ravel(), **fit_kwargs, sample_weight=sample_weight)


def predict(model, dataset_iterator: WindowedDatasetIterator):
    """
    Get the predictions for the given iterator

    Parameters:
        dataset_iterator: Dataset iterator from which obtain data to predict

    Returns:
       Array with the predictiosn
    """
    X, _, _ = dataset_iterator.get_data()
    return model.predict(X)


def fit_batch(
    model, train_batcher: Batcher, val_batcher: Batcher, n_epochs=15
) -> Tuple["Model", List]:
    """
    Fit the model using the given batcher

    Parameters:
        model: SKLearn Model
        train_batcher: Train dataset batcher
        val_batcher: Validation dataset batcher
        n_epochs: Number of epochs, by default 15

    Returns:
        model: the model
        history: history of errors
    """

    history = []
    for r in range(n_epochs):
        for X, y in train_batcher:
            model.partial_fit(np.reshape(X, (X.shape[0], X.shape[1] * X.shape[2])), y)
        y_true, y_pred = predict(val_batcher)
        history.append(mean_squared_error(y_true, y_pred))
    return model, history


def predict_batch(model, dataset_batcher: Batcher) -> np.ndarray:
    """
    Predict the values using the given batcher

    Parameters:
        model: SKLearn model
        dataset_batcher: The batcher


    Returns:
        RUL Prediction array
    """
    y_pred = []
    for X, y in dataset_batcher:
        y_pred.extend(
            np.squeeze(
                model.predict(np.reshape(X, (X.shape[0], X.shape[1] * X.shape[2])))
            )
        )
    return np.squeeze(np.array(y_pred))




from typing import List

import tensorflow as tf
from ceruleo.models.keras.dataset import KerasTrainableModel
from ceruleo.models.keras.losses import time_to_failure_rul
from tensorflow.keras import Input, Model, optimizers
from tensorflow.keras.layers import LSTM, Dense


class MultiTaskRUL(KerasTrainableModel):
    """
        A Multi task network that learns to regress the RUL and the Time to failure

        Two Birds with One Network: Unifying Failure Event Prediction and Time-to-failure Modeling
        Karan Aggarwal, Onur Atan, Ahmed K. Farahat, Chi Zhang, Kosta Ristovski, Chetan Gupta

        The target

        Parameters
        -----------
        layers_lstm : List[int]
                      Number of LSTM layers
        layers_dense : List[int]
                       Number of dense layers
        window: int

        batch_size: int
        step: int
        transformer
        shuffle
        models_path
        patience: int = 4
        cache_size: int = 30
    """

    def __init__(self,
                 layers_lstm: List[int],
                 layers_dense: List[int],
                 window: int,
                 batch_size: int,
                 step: int, transformer, shuffle, models_path,
                 patience: int = 4, cache_size: int = 30):
        super().__init__(window,
                         batch_size,
                         step,
                         transformer,
                         shuffle,
                         models_path,
                         patience=4,
                         cache_size=30)
        self.layers_dense = layers_dense
        self.layers_lstm = layers_lstm

    def compile(self):
        super().compile()
        self.model.compile(
            optimizer=optimizers.Adam(lr=0.001),
            loss=time_to_failure_rul(weights={
                0: 1.,
                1: 2.
            }),
            # {
            #    'rul': MeanSquaredError(),
            #    'ttf': BinaryCrossentropy(from_logits=True),
            # },
            loss_weights=[1.0, 1.0],
        )

    @property
    def name(self):
        return "MultiTaskRULTTF"

    def build_model(self):
        n_features = self.transformer.n_features

        input = Input(shape=(self.window, n_features))
        x = input
        n_filters = 15
        if len(self.layers_lstm) > 1:
            for n_filters in self.layers_lstm:
                x = LSTM(n_filters, recurrent_dropout=0.2,
                         return_sequences=True)(x)

        x = LSTM(n_filters, recurrent_dropout=0.2, return_sequences=False)(x)

        for n_filters in self.layers_dense:
            x = Dense(n_filters, activation='elu')(x)

        RUL_output = Dense(1, activation='elu', name='rul')(x)

        FP_output = Dense(1, activation='sigmoid', name='ttf')(x)

        output = tf.keras.layers.Concatenate(axis=1)([RUL_output, FP_output])
        model = Model(
            inputs=[input],
            outputs=[output],
        )
        return model

    def _generate_batcher(self, train_batcher, val_batcher):
        n_features = self.transformer.n_features

        def gen_train():
            for X, y in train_batcher:
                yield X, y

        def gen_val():
            for X, y in val_batcher:
                yield X, y

        a = tf.data.Dataset.from_generator(
            gen_train, (tf.float32, tf.float32), (tf.TensorShape(
                [None, self.window, n_features]), tf.TensorShape([None, 2])))
        b = tf.data.Dataset.from_generator(
            gen_val, (tf.float32, tf.float32), (tf.TensorShape(
                [None, self.window, n_features]), tf.TensorShape([None, 2])))
        return a, b



"""Compute evaluating results of fitted models

One of the most important issues regarding PM is the ability to compare and evaluate different methods.

The main data structure used in the results module is a dictionary in which each of the keys is the model name, 
and the elements are a list of PredictionResult. 
Each element of the model array is interpreted as a Fold in CV settings. 

Additionally to the regression error, it is possible to compute some metrics more easily interpretable. In this context, two metrics were defined in [1], namely:

- Frequency of Unexpected Breaks (ρUB) - the percentage of failures not prevented;
- Amount of Unexploited Lifetime (ρUL) - the average number of time that could have been run before failure if the preventative maintenance suggested by the maintenance management mod-ule had not been performed.

When compaing between the predicted end of life with respect to the true end of that particular life three scenarios can happen:

- The predicted end of life occurs before the true one. In that case, the predictions were pessimistic and the tool could have been used more time.
- The remaining useful life arrives at zero after the true remaining useful life. In that case, we incur the risk of the tool breaking.  
- The predicted line coincides with the true line. In that case, we don’t have unexploited time, and the risk of breakage can be considered 0.

Since usually the breakages are considered more harmful, a possible approach to preventing unexpected failures is to consider a more conservative maintenance approach, providing maintenance tasks recommendations some time before the end of life predicted. In that way, a conservative window can be defined in which the recommendation of making maintenance task should be performed at time T-predicted - conservative window size.

[1] Machine Learning for Predictive Maintenance: A Multiple Classifiers Approach
    Susto, G. A., Schirru, A., Pampuri, S., McLoone, S., & Beghi, A. (2015). 


"""
import logging
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from sklearn.metrics import mean_absolute_error as mae
from sklearn.metrics import mean_absolute_percentage_error as mape
from sklearn.metrics import mean_squared_error as mse
from uncertainties import ufloat

from ceruleo.results.picewise_regression import (
    PiecewesieLinearFunction,
    PiecewiseLinearRegression,
)

logger = logging.getLogger(__name__)


@dataclass
class MetricsResult:
    """An object that store regression metrics and times"""

    mae: float
    mse: float
    fitting_time: float = field(default_factory=lambda: 0)
    prediction_time: float = field(default_factory=lambda: 0)


@dataclass
class PredictionResult:
    """A prediction result is composed by a name"""

    name: str
    true_RUL: np.ndarray
    predicted_RUL: np.ndarray
    metrics: MetricsResult = field(default_factory=lambda: MetricsResult(0, 0))

    def compute_metrics(self):
        self.metrics.mae = mae(self.true_RUL, self.predicted_RUL)
        self.metrics.mse = mse(self.true_RUL, self.predicted_RUL)

    def __post_init__(self):    
        self.true_RUL = np.squeeze(self.true_RUL)
        self.predicted_RUL = np.squeeze(self.predicted_RUL)
        self.compute_metrics()


def compute_sample_weight(sample_weight, y_true, y_pred, c: float = 0.9):
    if sample_weight == "relative":
        sample_weight = np.abs(y_true - y_pred) / (np.clip(y_true, c, np.inf))
    else:
        sample_weight = 1
    return sample_weight


def compute_rul_line(rul: float, n: int, tt: Optional[np.array] = None):
    if tt is None:
        tt = -np.ones(n)
    z = np.zeros(n)
    z[0] = rul
    for i in range(len(tt) - 1):
        z[i + 1] = max(z[i] + tt[i], 0)
        if z[i + 1] - 0 < 0.0000000000001:
            break
    return z


class CVResults:
    """
    Compute the error histogram

    Compute the error with respect to the RUL considering the results of different
    folds

    Parameters:
        y_true: List with the true values of each hold-out set of a cross validation
        y_pred: List with the predictions of each hold-out set of a cross validation
        nbins: Number of bins to compute the histogram

    """

    def __init__(
        self,
        y_true: List[List],
        y_pred: List[List],
        nbins: int = 5,
        bin_edges: Optional[np.array] = None,
    ):
        if bin_edges is None:
            max_value = np.max([np.max(y) for y in y_true])
            bin_edges = np.linspace(0, max_value, nbins + 1)
        self.n_folds = len(y_true)
        self.n_bins = len(bin_edges) - 1
        self.bin_edges = bin_edges
        self.mean_error = np.zeros((self.n_folds, self.n_bins))
        self.mae = np.zeros((self.n_folds, self.n_bins))
        self.mse = np.zeros((self.n_folds, self.n_bins))
        self.errors = []
        for i, (y_pred, y_true) in enumerate(zip(y_pred, y_true)):
            self._add_fold_result(i, y_pred, y_true)

    def _add_fold_result(self, fold: int, y_pred: np.array, y_true: np.array):
        y_pred = np.squeeze(y_pred)
        y_true = np.squeeze(y_true)

        for j in range(len(self.bin_edges) - 1):
            mask = (y_true >= self.bin_edges[j]) & (y_true <= self.bin_edges[j + 1])
            indices = np.where(mask)[0]

            if len(indices) == 0:
                continue

            errors = y_true[indices] - y_pred[indices]

            self.mean_error[fold, j] = np.mean(errors)

            self.mae[fold, j] = np.mean(np.abs(errors))
            self.mse[fold, j] = np.mean((errors) ** 2)
            self.errors.append(errors)


def model_cv_results(
    results: List[PredictionResult],
    nbins: Optional[int] = None,
    bin_edges: Optional[np.ndarray] = None,
) -> CVResults:
    if nbins is None and bin_edges is None:
        raise ValueError("nbins and bin_edges cannot be both None")
    if nbins is None:
        nbins = len(bin_edges) - 1
    if bin_edges is None:
        max_y_value = np.max([r.true_RUL.max() for r in results])
        bin_edges = np.linspace(0, max_y_value, nbins + 1)

    trues = []
    predicted = []
    for results in results:
        trues.append(results.true_RUL)
        predicted.append(results.predicted_RUL)
    return CVResults(trues, predicted, nbins=nbins, bin_edges=bin_edges)


def models_cv_results(
    results_dict: Dict[str, List[PredictionResult]], nbins: int
) -> Tuple[np.ndarray, Dict[str, CVResults]]:
    """Create a dictionary with the result of each cross validation of the model"""
    max_y_value = np.max(
        [
            r.true_RUL.max()
            for model_name in results_dict.keys()
            for r in results_dict[model_name]
        ]
    )
    bin_edges = np.linspace(0, max_y_value, nbins + 1)
    model_results = {}
    for model_name in results_dict.keys():
        model_results[model_name] = model_cv_results(
            results_dict[model_name], bin_edges=bin_edges
        )

    return bin_edges, model_results


class FittedLife:
    """Represent a Fitted run-to-cycle failure

    Parameters:
        y_true: The true RUL target
        y_pred: The predicted target
        time: Time feature
        fit_line_not_increasing: Wether the fitted line can increase or not.
        RUL_threshold: Indicates the thresholding value used during  the fit

    """

    def __init__(
        self,
        y_true: np.array,
        y_pred: np.array,
        time: Optional[Union[np.array, int]] = None,
        fit_line_not_increasing: Optional[bool] = False,
        RUL_threshold: Optional[float] = None,
    ):
        self.fit_line_not_increasing = fit_line_not_increasing
        y_true = np.squeeze(y_true)
        y_pred = np.squeeze(y_pred)

        if time is not None:
            self.degrading_start = FittedLife._degrading_start(y_true, RUL_threshold)
            if isinstance(time, np.ndarray):
                self.time = time
            else:
                self.time = np.array(np.linspace(0, y_true[0], n=len(y_true)))

        else:
            self.degrading_start, self.time = FittedLife.compute_time_feature(
                y_true, RUL_threshold
            )

        # self.y_pred_fitted_picewise = self._fit_picewise_linear_regression(y_pred)
        # self.y_true_fitted_picewise = self._fit_picewise_linear_regression(y_true)

        self.RUL_threshold = RUL_threshold
        self.y_pred = y_pred
        self.y_true = y_true

        self.y_pred_fitted_coefficients = np.polyfit(self.time, self.y_pred, 1)
        p = np.poly1d(self.y_pred_fitted_coefficients)
        self.y_pred_fitted = p(self.time)

        self.y_true_fitted_coefficients = np.polyfit(self.time, self.y_true, 1)
        p = np.poly1d(self.y_true_fitted_coefficients)
        self.y_true_fitted = p(self.time)

    @staticmethod
    def compute_time_feature(
        y_true: np.array, RUL_threshold: Optional[float] = None
    ) -> Tuple[float, np.ndarray]:
        """Compute the time feature based on the target

        Parameters:
            y_true: RUL target
            RUL_threshold:

        Returns:
            Degrading start time and time
        """
        degrading_start = FittedLife._degrading_start(y_true, RUL_threshold)
        time = FittedLife._compute_time(y_true, degrading_start)
        return degrading_start, time

    @staticmethod
    def _degrading_start(
        y_true: np.array, RUL_threshold: Optional[float] = None
    ) -> float:
        """
        Obtain the index when the life value is lower than the RUL_threshold

        Parameters:
            y_true: Array of true values of the RUL of the life
            RUL_threshold: float


        Return:
            If RUL_threshold is None, the degrading start if the first index.
            Otherwise it is the first index in which y_true < RUL_threshold
        """
        degrading_start = 0
        if RUL_threshold is not None:
            degrading_start_i = np.where(y_true < RUL_threshold)
            if len(degrading_start_i[0]) > 0:
                degrading_start = degrading_start_i[0][0]
        else:
            d = np.diff(y_true) == 0
            while (degrading_start < len(d)) and (d[degrading_start]):
                degrading_start += 1
        return degrading_start

    @staticmethod
    def _compute_time(y_true: np.array, degrading_start: int) -> np.array:
        """
        Compute the passage of time from the true RUL

        The passage of time is computed as the cumulative sum of the first
        difference of the true labels. In case there are tresholded values,
        the time steps of the thresholded zone is assumed to be as the median values
        of the time steps computed in the zones of the life in which we have information.

        Parameters:
            y_true: The true RUL labels
            degrading_start: The index in which the true RUL values starts to be lower than the treshold

        Returns:
            Time component
        """
        if len(y_true) == 1:
            return np.array([0])

        time_diff = np.diff(np.squeeze(y_true)[degrading_start:][::-1])
        time = np.zeros(len(y_true))
        if degrading_start > 0:
            if len(time_diff) > 0:
                time[0 : degrading_start + 1] = np.median(time_diff)
            else:
                time[0 : degrading_start + 1] = 1
        time[degrading_start + 1 :] = time_diff

        return np.cumsum(time)

    def _fit_picewise_linear_regression(self, y: np.array) -> PiecewesieLinearFunction:
        """
        Fit the array trough a picewise linear regression

        Parameters:
            y: Points to be fitted
        Returns:
            The Picewise linear function fitted
        """
        pwlr = PiecewiseLinearRegression(not_increasing=self.fit_line_not_increasing)
        for j in range(len(y)):
            pwlr.add_point(self.time[j], y[j])
        line = pwlr.finish()

        return line

    def rmse(self, sample_weight=None) -> float:
        N = len(self.y_pred)
        sw = compute_sample_weight(sample_weight, self.y_true[:N], self.y_pred)
        return np.sqrt(np.mean(sw * (self.y_true[:N] - self.y_pred) ** 2))

    def mae(self, sample_weight=None) -> float:
        N = len(self.y_pred)
        sw = compute_sample_weight(sample_weight, self.y_true[:N], self.y_pred)
        return np.mean(sw * np.abs(self.y_true[:N] - self.y_pred))

    def noisiness(self) -> float:
        """
        How much the predictions resembles a line

        This metric is computed as the mse of the fitted values
        with respect to the least squares fitted line of this
        values

        Returns:
            The Mean Absolute Error of the fitted values with respect to the least squares fitted line
        """
        return mae(self.y_pred_fitted, self.y_pred)

    def slope_resemblance(self):
        m1 = self.y_true_fitted_coefficients[0]
        m2 = self.y_pred_fitted_coefficients[0]
        d = np.arctan((m1 - m2) / (1 + m1 * m2))
        d = d / (np.pi / 2)
        return 1 - np.abs((d / (np.pi / 2)))

    def predicted_end_of_life(self):
        z = np.where(self.y_pred == 0)[0]
        if len(z) == 0:
            return self.time[len(self.y_pred) - 1] + self.y_pred[-1]
        else:
            return self.time[z[0]]

    def end_of_life(self):
        z = np.where(self.y_true == 0)[0]
        if len(z) == 0:
            return self.time[len(self.y_pred) - 1] + self.y_true[-1]
        else:
            return self.time[z[0]]

    def maintenance_point(self, m: float = 0) -> float:
        """
        Compute the maintenance point

        The maintenance point is computed as the predicted end of life - m

        Parameters:
            m: Fault horizon  Defaults to 0.

        Returns:
            Time of maintenance
        """
        return self.predicted_end_of_life() - m

    def unexploited_lifetime(self, m: float = 0) -> float:
        """
        Compute the unexploited lifetime given a fault horizon window

        Machine Learning for Predictive Maintenance: A Multiple Classifiers Approach
        Susto, G. A., Schirru, A., Pampuri, S., McLoone, S., & Beghi, A. (2015).

        Parameters:
            m: Fault horizon window. Defaults to 0.

        Returns:
            Unexploited lifetime
        """

        if self.maintenance_point(m) < self.end_of_life():
            return self.end_of_life() - self.maintenance_point(m)
        else:
            return 0

    def unexpected_break(self, m: float = 0, tolerance: float = 0) -> bool:
        """
        Compute weather an unexpected break will produce using a fault horizon window of size m

        Machine Learning for Predictive Maintenance: A Multiple Classifiers Approach
        Susto, G. A., Schirru, A., Pampuri, S., McLoone, S., & Beghi, A. (2015).

        Parameters:
            m: Fault horizon window.

        Returns:
            A boolean indicating if an unexpected break will occur
        """
        if self.maintenance_point(m) - tolerance < self.end_of_life():
            return False
        else:
            return True


def split_lives_indices(y_true: np.array) -> List[List[int]]:
    """
    Obtain a list of indices for each life

    Parameters:
        y_true: True vector with the RUL

    Returns:
        A list with the indices belonging to each life
    """
    assert len(y_true) >= 2
    lives_indices = (
        [0]
        + (np.where(np.diff(np.squeeze(y_true)) > 0)[0] + 1).tolist()
        + [len(y_true)]
    )
    indices = []
    for i in range(len(lives_indices) - 1):
        r = range(lives_indices[i], lives_indices[i + 1])
        if len(r) <= 1:
            continue
        indices.append(r)
    return indices


def split_lives(
    results: PredictionResult,
    RUL_threshold: Optional[float] = None,
    fit_line_not_increasing: Optional[bool] = False,
    time: Optional[int] = None,
) -> List[FittedLife]:
    """
    Divide an array of predictions into a list of FittedLife Object

    Parameters:
        y_true: The true RUL target
        y_pred: The predicted RUL
        fit_line_not_increasing: Weather the fit line can increase, by default False
        time:  A vector with timestamps. If omitted will be computed from y_true

    Returns:
       FittedLife list
    """
    lives = []
    for r in split_lives_indices(results.true_RUL):
        if np.any(np.isnan(results.predicted_RUL[r])):
            continue
        lives.append(
            FittedLife(
                results.true_RUL[r],
                results.predicted_RUL[r],
                RUL_threshold=RUL_threshold,
                fit_line_not_increasing=fit_line_not_increasing,
                time=time,
            )
        )
    return lives


def unexploited_lifetime(d: PredictionResult, window_size: int, step: int):
    bb = [split_lives(cv) for cv in d]
    return unexploited_lifetime_from_cv(bb, window_size, step)


def unexploited_lifetime_from_cv(
    lives: List[List[FittedLife]], window_size: int, n: int
):
    std_per_window = []
    mean_per_window = []
    windows = np.linspace(0, window_size, n)
    for m in windows:
        jj = []
        for r in lives:
            ul_cv_list = [life.unexploited_lifetime(m) for life in r]

            jj.extend(ul_cv_list)
        mean_per_window.append(np.mean(jj))
        std_per_window.append(np.std(jj))

    return windows, np.array(mean_per_window), np.array(std_per_window)


def unexpected_breaks(
    d: List[PredictionResult], window_size: int, step: int
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the risk of unexpected breaks with respect to the maintenance window size

    Parameters:
        d: Dictionary with the results
        window_size: Maximum size of the maintenance windows
        step: Number of points in which compute the risks.
            step different maintenance windows will be used.

    Returns:
        A tuple of np.arrays with:
            - Maintenance window size evaluated
            - Risk computed for every window size used
    """

    bb = [split_lives(fold) for fold in d]
    return unexpected_breaks_from_cv(bb, window_size, step)


def unexpected_breaks_from_cv(
    lives: List[List[FittedLife]], window_size: int, n: int
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the risk of unexpected breaks given a Cross-Validation results

    Parameters:
        lives: Cross validation results.
        window_size: Maximum size of the maintenance window
        n: Number of points to evaluate the risk of unexpected breaks


    Returns:
        A tuple of np.arrays with:
            - Maintenance window size evaluated
            - Risk computed for every window size used
    """
    std_per_window = []
    mean_per_window = []
    windows = np.linspace(0, window_size, n)
    for m in windows:
        jj = []
        for r in lives:
            ul_cv_list = [life.unexpected_break(m) for life in r]
            jj.extend(ul_cv_list)
        mean_per_window.append(np.mean(jj))
        std_per_window.append(np.std(jj))
    return windows, np.array(mean_per_window), np.array(std_per_window)


def metric_J_from_cv(lives: List[List[FittedLife]], window_size: int, n: int, q1, q2):
    J = []
    windows = np.linspace(0, window_size, n)
    for m in windows:
        J_of_m = []
        for r in lives:
            ub_cv_list = np.array([life.unexpected_break(m) for life in r])
            ub_cv_list = (ub_cv_list / (np.max(ub_cv_list) + 0.0000000001)) * q1
            ul_cv_list = np.array([life.unexploited_lifetime(m) for life in r])
            ul_cv_list = (ul_cv_list / (np.max(ul_cv_list) + 0.0000000001)) * q2
            values = ub_cv_list + ul_cv_list
            mean_J = np.mean(values)
            std_ul_cv = np.std(values)
            J_of_m.append(mean_J)
        J.append(np.mean(J_of_m))
    return windows, J


def metric_J(d, window_size: int, step: int):
    lives_cv = [split_lives(cv) for cv in d]
    return metric_J_from_cv(lives_cv, window_size, step)


def cv_regression_metrics_single_model(
    results: List[PredictionResult], threshold: float = np.inf
):
    errors = {"MAE": [], "MAE SW": [], "MSE": [], "MSE SW": [], "MAPE": []}
    for result in results:
        y_mask = np.where(result.true_RUL <= threshold)[0]
        y_true = np.squeeze(result.true_RUL[y_mask])
        y_pred = np.squeeze(result.predicted_RUL[y_mask])
        mask = np.isfinite(y_pred)
        y_pred = y_pred[mask]
        y_true = y_true[mask]

        if len(np.unique(y_pred)) == 1:
            continue

        sw = compute_sample_weight(
            "relative",
            y_true,
            y_pred,
        )
        try:
            MAE_SW = mae(
                y_true,
                y_pred,
                sample_weight=sw,
            )
        except:
            MAE_SW = np.nan
        try:
            MAE = mae(y_true, y_pred)
        except:
            MAE = np.nan

        try:
            MSE_SW = mse(
                y_true,
                y_pred,
                sample_weight=sw,
            )
        except:
            MSE_SW = np.nan

        try:
            MSE = mse(y_true, y_pred)
        except:
            MSE = np.nan

        try:
            MAPE = mape(y_true, y_pred)
        except:
            MAPE = np.nan

        lives = split_lives(result)
        errors["MAE"].append(MAE)
        errors["MAE SW"].append(MAE_SW)
        errors["MSE"].append(MSE)
        errors["MSE SW"].append(MSE_SW)
        errors["MAPE"].append(MAPE)

    errors1 = {}
    for k in errors.keys():
        errors1[k] = ufloat(
            np.round(np.mean(errors[k]), 2), np.round(np.std(errors[k]), 2)
        )
    return errors1


def cv_regression_metrics(
    results_dict: Dict[str, List[PredictionResult]], threshold: float = np.inf
) -> dict:
    """
    Compute regression metrics for each model

    Parameters:
<<<<<<< HEAD

        data: Dictionary with the model predictions.

=======
        data: Dictionary with the model predictions.
>>>>>>> 7862b3c56dbc36f72b3c7f87d9b39e1ae78b4ddc
        threshold: Compute metrics errors only in RUL values less than the threshold

    Returns:
        A dictionary with the following structure:
            d: { ['Model]:
                    {
                        'MAE': {
                            'mean':
                            'std':
                        },
                        'MAE SW': {
                            'mean':
                            'std':
                        },
                        'MSE': {
                            'mean':
                            'std':
                        },
                    }
                ]

<<<<<<< HEAD

        d: { ['Model]:
                {
                    'MAE': {
                        'mean':
                        'std':
                    },
                    'MAE SW': {
                        'mean':
                        'std':
                    },
                    'MSE': {
                        'mean':
                        'std':
                    },
                }
            ]

=======
>>>>>>> 7862b3c56dbc36f72b3c7f87d9b39e1ae78b4ddc
    """
    out = {}
    for model_name in results_dict.keys():
        out[model_name] = cv_regression_metrics_single_model(
            results_dict[model_name], threshold
        )
    return out





import numpy as np
import pandas as pd

from ceruleo.transformation.features.scalers import (RobustMinMaxScaler)
from sklearn.preprocessing import RobustScaler

class TestImputers():

    def test_RobustMinMaxScaler(self):

        scaler = RobustMinMaxScaler(range=(-1, 1), clip=False, lower_quantile=0.1, upper_quantile=0.9)
        sk_scaler = RobustScaler(with_centering=False, with_scaling=True, quantile_range=(10, 90))

        df1 = pd.DataFrame({
            'a': np.random.randn(6000)*5 + 25,
            'b': np.random.randn(6000)*5 + 25
        })
        scaler.partial_fit(df1)


        df2 = pd.DataFrame({
            'a': np.random.randn(6000)*3 + 15,
            'b': np.random.randn(6000)*3 + 88
        })
        scaler.partial_fit(df2)


        sk_scaler.fit(pd.concat((df1, df2)))

        sk_scaler.transform(df1)
      
  



from pathlib import Path
import os

PACKAGE_PATH = Path(__file__).resolve().parent
DATA_PATH = Path(os.environ.get('CERULEO_DATA_PATH', Path.home() / '.ceruleo' / 'data'))
DATA_PATH.mkdir(parents=True, exist_ok=True)

CACHE_PATH = Path(os.environ.get('CERULEO_CACHE_PATH', Path.home() / '.ceruleo' / 'cache'))
CACHE_PATH.mkdir(parents=True, exist_ok=True)


__version__ = "3.0.6"



from ceruleo.transformation.functional.graph_utils import topological_sort_iterator
from ceruleo.transformation.functional.transformerstep import TransformerStep


class VisitableNode(TransformerStep):
    def visit(self):
        pass


class A(VisitableNode):
    pass


class B(VisitableNode):
    class B1(VisitableNode):
        pass

    class B2(VisitableNode):
        pass

    def __init__(self, *, name):
        super().__init__(name=name)
        self.b1 = B.B1(name="B1")
        self.b2 = B.B2(name="B2")(self.b1)


class VisitableNode(TransformerStep):
    def visit(self):
        pass


class A(VisitableNode):
    pass


class B(VisitableNode):
    class B1(VisitableNode):
        pass

    class B2(VisitableNode):
        pass

    def __init__(self, *, name: str):
        super().__init__(name=name)
        self.b1 = B.B1(name="B1")
        self.b2 = B.B2(name="B2")(self.b1)

    def visit(self):
        for n in self.next:
            self.disconnect(n)
            self.b2.add_next(n)
        self.b1(self)


class C(VisitableNode):
    pass


class Node(TransformerStep):
    pass

    def visit(self):
        self.disconnect(self.next)
        self.b2.add_next(self.next)
        self.b1(self)


class C(VisitableNode):
    pass


class Node(TransformerStep):
    pass


class TestGraph:
    def test_simple(self):
        pipe = Node(name="A")
        pipe = Node(name="B")(pipe)
        pipe = Node(name="C")(pipe)

        assert pipe.previous[0].name == "B"
        assert pipe.previous[0].previous[0].name == "A"

        topological_sort_iterator

    def test_graph_updating(self):
        pipe = A(name="A")
        pipe = B(name="B")(pipe)
        pipe = C(name="C")(pipe)

        pipe.previous[0].visit()
        assert pipe.previous[0].name == "B2"
        assert pipe.previous[0].previous[0].name == "B1"
        assert pipe.previous[0].previous[0].previous[0].name == "B"

        pipe = A(name="A")
        pipe = B(name="B")(pipe)
        pipe = C(name="C")(pipe)

        result = []
        for a in topological_sort_iterator(pipe):
            a.visit()
            result.append(a.name)

        assert result == ["A", "B", "B1", "B2", "C"]

    def test_diamond(self):
        pipe = Node(name="A")
        pipeB = Node(name="B")(pipe)
        pipeC = Node(name="C")(pipe)
        pipeD = Node(name="D")(pipe)
        pipe = Node(name="E")([pipeB, pipeC, pipeD])




from ceruleo.utils.lrucache import LRUDataCache


class TestLRUCache():

    def test_limit_size(self):
        cache = LRUDataCache(3)
        cache.add('A', 5)
        cache.add('B', 6)
        cache.add('C', 7)
        cache.add('D', 5)
        assert len(cache) == 3

    def test_lru(self):
        cache = LRUDataCache(3)
        cache.add('A', 5)
        cache.add('B', 6)
        cache.add('C', 7)
        assert len(cache) == 3
        cache.get('A')
        cache.get('A')
        cache.get('A')
        cache.get('B')
        assert cache.data['A']['hit'] == 3
        assert cache.data['B']['hit'] == 1
        cache.add('D', 8)
        assert len(cache) == 3
        assert 'D' in cache.data
        assert 'C' not in cache.data
        cache.get('D')
        cache.get('D')
        cache.add('E', 9)
        assert 'B' not in cache.data



from typing import List, Tuple

import tensorflow as tf
from tensorflow.keras import Input, Sequential
from tensorflow.keras.layers import (LSTM, Conv1D, Dense, Dropout, Flatten,
                                     MaxPool1D, Reshape)


def CNLSTM(
    input_shape: Tuple[int, int],
    *,
    n_conv_layers: int,
    initial_convolutional_size: int,
    layers_recurrent: List[int],
    hidden_size: Tuple[int, int],
    dense_layer_size: int = 50,
    dropout: float = 0.1,
):
    """
    The network contains stacked layers of 1-dimensional convolutional layers
    followed by max poolings

    * Temporal Convolutional Memory Networks forRemaining Useful Life Estimation of Industrial Machinery
    * Lahiru Jayasinghe, Tharaka Samarasinghe, Chau Yuen, Jenny Chen Ni Low, Shuzhi Sam Ge

    [Reference](https://ieeexplore.ieee.org/abstract/document/8754956?casa_token=B_YvavFGulsAAAAA:f2k2I8pH1lM3sOcSGlXEF29seYPK1GPa9Od2-TwnhNeFyWvRRUAqkUOdWUNIyy9FPJHhsGM)

    Parameters:

        input_shape: Input shape of the iterator
        n_conv_layers: Number of convolutional layers. Each convolutional layers is composed by:
            a 1D-convolution:  kernelsize=2, strides=1,padding=same, activation=ReLu
            and a 1D-max-pooling   poolsize=2, strides=2, padding=same
        initial_convolutional_size: The number of filters of the first convolutional layers.
              Next ones will have the power of 2 of the previous one
        layers_recurrent: Number of current layers. Each recurrent layer is composed by an LSTM layer
        hidden_size: After the convolutional layers the signal is projected via a RELU layer and then
                     reshaped again as a time series of size (hidden_size[0], hidden_size[1])
        dropout: Droput factor
        dense_layer_size: Size of the dense layer before the head
    """

    model = Sequential()
    model.add(Input(shape=input_shape))

    for n_filters in range(n_conv_layers):
        model.add(
            Conv1D(
                filters=initial_convolutional_size,
                strides=1,
                kernel_size=2,
                padding="same",
                activation="relu",
            )
        )
        model.add(MaxPool1D(pool_size=2, strides=2))
        initial_convolutional_size = initial_convolutional_size * 2

    model.add(Flatten())
    model.add(Dense(hidden_size[0] * hidden_size[1], activation="relu"))
    model.add(Dropout(dropout))
    model.add(Reshape(hidden_size))

    for i, n_filters in enumerate(layers_recurrent):

        model.add(
            LSTM(
                n_filters, return_sequences=i < len(layers_recurrent) - 1
            )
        )

    model.add(Dropout(dropout))

    model.add(Flatten())
    model.add(Dense(dense_layer_size, activation="relu"))
    model.add(Dropout(dropout))
    model.add(Dense(1, activation="relu"))
    return model



from typing import Union

from ceruleo.dataset.transformed import TransformedDataset
from ceruleo.dataset.ts_dataset import AbstractPDMDataset


def iterate_over_features(ds: Union[TransformedDataset, AbstractPDMDataset]):
    """Helper function to iterate over the features in a dataset
    Example:
        for X, y, metadata in df:
            pass
        for X in iterate_over_features(ds):
            pass
    Parameters:
        ds: The dataset
    Returns:

        it: The iterator
    """
    if isinstance(ds, TransformedDataset):
        return map(lambda x: x[0], ds)
    else:
        return ds


def iterate_over_target(ds: Union[TransformedDataset, AbstractPDMDataset]):
    """Helper function to iterate over the RUL target in a dataset
    Example:

        for X, y, metadata in df:
            pass

        for y in iterate_over_target(ds):
            pass

    Parameters:
        ds: The dataset

    Returns:

        it: The iterator
    """
    if isinstance(ds, TransformedDataset):
        return map(lambda x: x[1], ds)
    elif hasattr(ds, "rul_column"):
        return map(lambda x: x[ds.rul_column], ds)
    else:
        raise ValueError("Invalid dataset type used")


def iterate_over_features_and_target(ds: Union[TransformedDataset, AbstractPDMDataset]):
    """Helper function to iterate over the features and RUL target in a dataset
    Example:

        for X, y, metadata in df:
            pass

        for X, yy in iterate_over_features_and_target(ds):
            pass

    Parameters:
        ds: The dataset

    Returns:

        it: The iterator
    """
    if isinstance(ds, TransformedDataset):
        return map(lambda x: (x[0], x[1]), ds)
    elif hasattr(ds, "rul_column"):
        return map(lambda x: (x, x[ds.rul_column]), ds)
    else:
        raise ValueError("Invalid dataset type used")



"""
Shufflers are helpers classes used by the iterator to change
the order of the run-to-failure cycles and the timestamps inside
each cycle

There are six types of shuffles

- NotShuffled: All the cycles are processed in order
- AllShuffled: Everything is shuffled
- IntraTimeSeriesShuffler: Each point of the time series is shuffled, but the TS are kept in order
- InverseOrder: The data points will be fed in RUL decreasing order
- TimeSeriesOrderIntraSignalShuffling: Each point in the ts is shuffled, and the ts order are shuffled also.
- TimeSeriesOrderShuffling: Time series are shuffled, but each point inside the time series kept  its order
"""
from typing import Callable, Optional, Tuple

import numpy as np
import math


class AbstractShuffler:
    """
    A Shuffler is used by the iterator to interleave samples of different run-to-fail cycles
    """

    class Iterator:
        def __init__(self, shuffler, iterator):
            self.iterator = iterator
            self.shuffler = shuffler

        def __iter__(self):
            self.shuffler.initialize(self.iterator)
            return self

        def __next__(self):
            return self.shuffler.next_element()

    def iterator(self, iterator: "WindowedDatasetIterator"):
        return AbstractShuffler.Iterator(self, iterator)

    def at_end(self) -> bool:
        """
        Determines weather the iterator reached to its end

        Returns:
            bool: Wether the iterator is at its end
        """
        return self.current_time_series == self.wditerator.dataset.n_time_series

    def next_element(self) -> Tuple[int, int]:
        """
        Iterating function

        The method takes the current time serie, and the current time stamp
        and calls the advance method

        Returns:
            Index of the current run-to-failure cycle andcurrent timestamp of that cycle

        Raises:
           When the iteration reaches the end
        """

        if self.at_end():
            raise StopIteration
        ts_index = self.time_series()
        timestamp = self.timestamp()
        self.advance()

        return ts_index, timestamp

    def start(self, iterator: "WindowedDatasetIterator"):
        """
        Start the shuffler given an iterator

        Parameters:
            iterator: An Windowed Iterator
        """
        self.initialize(iterator)

    def time_series(self) -> int:
        """
        Current time series

        Returns:
            Current Time Series
        """
        return self.current_time_series

    def initialize(self, iterator: "WindowedDatasetIterator"):
        """
        Initialize the current shuffler

        This method is in charge of initializing everything to allow the correct iteration

        Parameters:
            iterator: WindowedDatasetIterator

        """
        self.wditerator = iterator
        self._samples_per_time_series = (
            np.ones(self.wditerator.dataset.n_time_series, dtype=np.int64)
            * np.iinfo(np.int64).max
        )
        self._time_series_sizes = (
            np.ones(self.wditerator.dataset.n_time_series, dtype=np.int64)
            * np.iinfo(np.int64).max
        )
        self.current_time_series = 0

    def load_time_series(self, time_series_index: int):
        total_length = self.wditerator.dataset.number_of_samples_of_time_series(
            time_series_index
        )
        start_index = self.wditerator.start_index.get(total_length)
        end_index = self.wditerator.end_index.get(total_length)

        N = end_index - start_index
        samples_per_step = N / self.wditerator.step
        self._samples_per_time_series[time_series_index] = math.ceil(samples_per_step)
        last_part_missing = (
            math.ceil((N - 1) / self.wditerator.step) - ((N - 1) / self.wditerator.step)
            > 0
        )
        if self.wditerator.last_point and last_part_missing:
            self._samples_per_time_series[time_series_index] += 1

        self._time_series_sizes[time_series_index] = total_length

    def number_samples_of_time_series(self, time_series_index: int) -> int:
        if self._samples_per_time_series[time_series_index] == np.iinfo(np.int64).max:
            self.load_time_series(time_series_index)
        return self._samples_per_time_series[time_series_index]

    def number_of_samples_of_current_time_series(self) -> int:
        """
        Obtain the number of samples for the current time series

        Returns:
            Number of samples

        """

        return self.number_samples_of_time_series(self.current_time_series)

    def timestamp(self) -> int:
        """Obtain a timestamp for the current run-to-failure cycle"""
        raise NotImplementedError

    def time_series_size(self, time_series_index: int):
        if self._time_series_sizes[time_series_index] == np.iinfo(np.int64).max:
            self.load_time_series(time_series_index)
        return self._time_series_sizes[time_series_index]

    def current_time_series_size(self):
        return self.time_series_size(self.current_time_series)

    def advance(self):
        """Abstract method for obtain the new pair (time series, timestamp)"""
        raise NotImplementedError


class IntraTimeSeriesShuffler(AbstractShuffler):
    """
    Each point of the time series is shuffled, but the TS are kept in order

        Iteration 1: | TS 1 | TS 1 | TS 1 | TS 2 | TS 2 | TS 2
                     |   3  |  1   |  2   |   2  |   3  |   1
        Iteration 2: | TS 1 | TS 1 | TS 1 | TS 2 | TS 2 | TS 2
                     |   1  |  3   |  2   |   3  |   2  |   1
    """

    def time_series_changed(self):
        self.current_timestamp_index = 0
        if self.current_time_series == self.wditerator.dataset.n_time_series:
            return
        self.timestamps = np.arange(
            start=0,
            stop=self.current_time_series_size(),
            step=self.wditerator.step,
            dtype=np.int64,
        )
        if (
            self.wditerator.last_point
            and self.timestamps[-1] != self.current_time_series_size() - 1
        ):
            self.timestamps = np.append(
                self.timestamps, self.current_time_series_size() - 1
            )
        np.random.shuffle(self.timestamps)

    def initialize(self, iterator: "WindowedDatasetIterator"):
        super().initialize(iterator)
        self.current_time_series = 0
        self.time_series_changed()
        self.n_time_series = iterator.dataset.n_time_series

    def timestamp(self) -> int:
        ret = self.timestamps[self.current_timestamp_index]
        return ret

    def advance(self):
        self.current_timestamp_index += 1

        if (
            self.current_timestamp_index
            >= self.number_of_samples_of_current_time_series()
        ):
            self.current_time_series += 1
            self.time_series_changed()


class TimeSeriesOrderShuffling(AbstractShuffler):
    """
    Time series are shuffled, but each point inside the time series kept  its order

        Iteration 1: | TS 1 | TS 1 | TS 1 | TS 2 | TS 2 | tS 2 |
                     |   1  | 2    |  3   |   1  |   2  |   3  |
        Iteration 2: | TS 2 | TS 2 | TS 2 | TS 1 | TS 1 | TS 1 |
                     |   1  |  2   |  3   |   1  |   2  |   3  |
    """

    def initialize(self, iterator: "WindowedDatasetIterator"):
        super().initialize(iterator)
        self.available_time_series = np.arange(
            iterator.dataset.n_time_series, dtype=np.int64
        )
        np.random.shuffle(self.available_time_series)
        self.available_time_series_index = 0
        self.current_timestamp = 0
        self.current_time_series = self.available_time_series[
            self.available_time_series_index
        ]

    def timestamp(self) -> int:
        ret = self.current_timestamp

        return ret

    def advance(self):
        if self.current_timestamp == self.current_time_series_size() - 1:
            self.time_series_changed()
        else:
            self.current_timestamp += self.wditerator.step
            self.current_timestamp = min(
                self.current_timestamp, self.current_time_series_size() - 1
            )

    def time_series_changed(self):
        self.current_timestamp = 0
        self.available_time_series_index += 1
        if self.available_time_series_index < len(self.available_time_series):
            self.current_time_series = self.available_time_series[
                self.available_time_series_index
            ]
        else:
            self.current_time_series = len(self.available_time_series)


class TimeSeriesOrderIntraSignalShuffling(AbstractShuffler):
    """
     Each point in the ts is shuffled, and the ts order are shuffled also.

    !!! note

        Iteration 1: | TS 1 | TS 1 | TS 1 | TS 2 | TS 2 | TS 2
                     |   3  | 2    |  1   |   1  |   3  |   2
        Iteration 2: | TS 2 | TS 2 | TS 2 | TS 1 | TS 1 | TS 1
                     |   3 |  1   |  2   |   3  |   1  |   2

    """

    def initialize(self, iterator: "WindowedDatasetIterator"):
        super().initialize(iterator)
        self.available_time_series = np.arange(
            iterator.dataset.n_time_series, dtype=np.int64
        )
        np.random.shuffle(self.available_time_series)
        self.available_time_series_index = 0
        self.current_time_series = self.available_time_series[
            self.available_time_series_index
        ]

        self.available_time_stamps = np.arange(
            start=0,
            stop=self.number_of_samples_of_current_time_series(),
            step=self.wditerator.step,
            dtype=np.int64,
        )
        self.available_time_stamps_index = 0

    def timestamp(self) -> int:
        ret = self.available_time_stamps[self.available_time_stamps_index]

        return ret

    def advance(self):
        self.available_time_stamps_index += 1
        if self.available_time_stamps_index == len(self.available_time_stamps):
            self.time_series_changed()

    def time_series_changed(self):
        self.available_time_stamps_index = 0
        self.available_time_series_index += 1
        if self.available_time_series_index < len(self.available_time_series):
            self.current_time_series = self.available_time_series[
                self.available_time_series_index
            ]
            self.available_time_stamps = np.arange(
                start=0,
                stop=self.current_time_series_size(),
                step=self.wditerator.step,
                dtype=np.int64,
            )

        else:
            self.current_time_series = len(self.available_time_series)


class InverseOrder(AbstractShuffler):
    """
    The data points will be fed in RUL decreasing order

        Iteration 1: | TS 2 | TS 1 | TS 2 | TS 1 | TS 2 | TS 1
                     |   4  | 3    |  3   |   2  |   2  |   1
    """

    def initialize(self, iterator: "WindowedDatasetIterator"):
        super().initialize(iterator)
        self.sizes = np.array(
            [self.time_series_size(i) for i in range(iterator.dataset.n_time_series)],
            dtype=np.int64,
        )

    def time_series(self) -> int:
        self.current_time_series = np.argmax(self.sizes)
        return self.current_time_series

    def timestamp(self):
        ret = self.sizes[self.current_time_series] - 1

        return ret

    def advance(self):
        self.sizes[self.current_time_series] = np.clip(
            np.int64(self.sizes[self.current_time_series]) - self.wditerator.step,
            0,
            np.inf,
        )
        if np.sum(self.sizes) == 0:
            self.current_time_series = len(self.sizes)


class AllShuffled(AbstractShuffler):
    """
    Everything is shuffled

        Iteration 1: | TS 1 | TS 2 | TS 2 | TS 1 | TS 1 | TS 2
                     |   3  | 2    |  1   |   1  |   2  |   3
    """

    def initialize(
        self, iterator: "WindowedDatasetIterator", evenly_sampled: bool = True
    ):
        super().initialize(iterator)
        self.evenly_sampled = evenly_sampled
        self.timestamps_per_ts = {
            i: None for i in range(self.wditerator.dataset.n_time_series)
        }
        self.timestamps_per_ts_indices = np.array(
            [0 for i in range(self.wditerator.dataset.n_time_series)], dtype=np.int64
        )

    def time_series(self) -> int:
        l = np.random.randint(self.wditerator.dataset.n_time_series)
        while (
            self.number_samples_of_time_series(l) == self.timestamps_per_ts_indices[l]
        ):
            l = np.random.randint(self.wditerator.dataset.n_time_series)
        self.current_time_series = l

        return self.current_time_series

    def at_end(self) -> bool:
        return (self._samples_per_time_series == self.timestamps_per_ts_indices).all()

    def timestamp(self) -> int:
        return self.timestamps_per_ts[self.current_time_series][
            self.timestamps_per_ts_indices[self.current_time_series]
        ]

    def advance(self):
        self.timestamps_per_ts_indices[self.current_time_series] += 1

    def load_time_series(self, time_series_index: int):
        super().load_time_series(time_series_index)
        if self.evenly_sampled:
            total_length = self.time_series_size(time_series_index)
            final_index = self.wditerator.end_index.get(total_length)

            self.timestamps_per_ts[time_series_index] = np.arange(
                start=self.wditerator.start_index.get(total_length),
                stop=final_index,
                step=self.wditerator.step,
                dtype=np.int64,
            )
            if self.timestamps_per_ts[time_series_index][-1] != final_index - 1:
                self.timestamps_per_ts[time_series_index] = np.append(
                    self.timestamps_per_ts[time_series_index], final_index - 1
                )

        else:
            N = self.time_series_size(time_series_index)
            step = self.wditerator.step
            self.timestamps_per_ts[time_series_index] = N - np.logspace(
                0, 1, num=math.ceil(N / step), base=N
            ).astype("int")
        np.random.shuffle(self.timestamps_per_ts[time_series_index])


class NotShuffled(AbstractShuffler):
    """
    Nothing is shuffled. Each sample of each run-to-failure cycle are iterated in order

        Iteration 1: | Life 1 | Life 1  | Life 1 | Life 2 | Life 3 | Life 3
                     |   1    | 2       |  3     |   1    |   2    |   1
        Iteration 2: | Life 1 | Life 1  | Life 1 | Life 2 | Life 3 | Life 3
                     |   1    | 2       |  3     |   1    |   2    |   1
    """

    def initialize(self, iterator: "WindowedDatasetIterator"):
        super().initialize(iterator)
        self.current_time_series = 0

        self.current_timestamp = self.wditerator.start_index.get(
            self.current_time_series_size()
        )

    def time_series_changed(self):
        self.current_time_series += 1
        if not self.at_end():
            self.current_timestamp = self.wditerator.start_index.get(
                self.current_time_series_size()
            )

    def timestamp(self) -> int:
        return self.current_timestamp

    def advance(self):
        final_timestamp = self.wditerator.end_index.get(self.current_time_series_size())
        if self.current_timestamp == final_timestamp - 1:
            self.time_series_changed()
        else:
            self.current_timestamp += self.wditerator.step
            self.current_timestamp = min(self.current_timestamp, final_timestamp - 1)



from typing import Tuple
import numpy as np
import tensorflow as tf

from ceruleo.models.keras.layers import ExpandDimension, RemoveDimension
from keras import Input, Model
from keras.layers import (
    Activation,
    BatchNormalization,
    Concatenate,
    Conv1D,
    Conv2D,
    Dense,
    GlobalAveragePooling1D,
)


def XCM(input_shape: Tuple[int, int], *, n_filters: int = 128, filter_window: int = 7):
    """
    A modified version of the model presented in  XCM: An Explainable Convolutional Neural Network for Multivariate Time Series Classification

    The classification head was replaced by an regression head

    Fauvel, K., Lin, T., Masson, V., Fromont, É., & Termier, A. (2021). 
    Xcm: An explainable convolutional neural network for multivariate time series classification. Mathematics, 
    9(23), 3137.


    """

    input = Input(input_shape)
    x = input

    x2d = ExpandDimension()(x)
    x2d = Conv2D(n_filters, (input_shape[0], 1), padding="same", name="first_conv2d")(
        x2d
    )

    x2d = BatchNormalization()(x2d)

    x2d = Activation("relu")(x2d)
    model_fisrt_conv2d = Model(
        inputs=[input],
        outputs=[x2d],
    )
    x2d_input = Input(shape=x2d.shape[1:])

    x2d = Conv2D(
        1,
        (1, 1),
        padding="same",
        activation="relu",
    )(x2d_input)

    x2d = RemoveDimension(3)(x2d)

    input = Input(shape=(input_shape[0], input_shape[1]))
    x = input
    x1d = Conv1D(
        n_filters,
        filter_window,
        padding="same",
    )(x)

    x1d = BatchNormalization()(x1d)

    x1d = Activation("relu")(x1d)
    model_fisrt_conv1d = Model(
        inputs=[input],
        outputs=[x1d],
    )
    x1d_input = Input(shape=x1d.shape[1:])

    x1d = Conv1D(
        1,
        1,
        padding="same",
        activation="relu",
    )(x1d_input)

    x = Concatenate()([x2d, x1d])

    x = Conv1D(n_filters, filter_window, padding="same")(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    x = GlobalAveragePooling1D()(x)

    x = Dense(n_filters, activation="relu")(x)
    output = Dense(1, activation="relu", name="RUL_regressor")(x)

    model_regression = Model(inputs=[x2d_input, x1d_input], outputs=[output])
    model_input = Input(shape=(input_shape[0], input_shape[1]))
    conv2d = model_fisrt_conv2d(model_input)
    conv1d = model_fisrt_conv1d(model_input)
    output = model_regression([conv2d, conv1d])

    model = Model(
        inputs=[model_input],
        outputs=output,
    )
    return model,  (model_fisrt_conv1d, model_fisrt_conv2d, model_regression)


def explain( model_extras, input):
    model_fisrt_conv1d, model_fisrt_conv2d, model_regression = model_extras
    data_input = np.expand_dims(input, axis=0)
    with tf.GradientTape() as tape:
        first_conv1d_layer_output = model_fisrt_conv1d(data_input)
        first_conv2d_layer_output = model_fisrt_conv2d(data_input)
        tape.watch(first_conv2d_layer_output)

        output = model_regression(
            [first_conv2d_layer_output, first_conv1d_layer_output]
        )  #
        grads = tape.gradient(output, first_conv2d_layer_output)
        filter_weight = np.mean(grads, axis=(0, 1, 2))

    mmap = np.sum(np.squeeze(first_conv2d_layer_output) * filter_weight, axis=(2))

    with tf.GradientTape() as tape:
        first_conv2d_layer_output = model_fisrt_conv2d(data_input)
        first_conv1d_layer_output = model_fisrt_conv1d(data_input)
        tape.watch(first_conv1d_layer_output)

        output = model_regression(
            [first_conv2d_layer_output, first_conv1d_layer_output]
        )

    grads1d = tape.gradient(output, first_conv1d_layer_output)
    filter1d_weight = np.mean(grads1d, axis=(0, 1))
    v = np.sum(np.squeeze(first_conv1d_layer_output) * filter1d_weight, axis=1)
    return (mmap, v)



from typing import List, Tuple

import numpy as np


class Segment:
    def __init__(self, initial_point:Tuple[float, float], not_increasing:bool = False):
        self.n = 1
        self.initial = initial_point
        self.xx = 0
        self.xy = 0
        self.yy = 0
        self.B = 0
        self.segment_error = []
        self.std_deviation = 2
        self.not_increasing = not_increasing

    def compute_endpoint(self, current_t: float):
        # Predict with previous point
        t = current_t - self.initial[0]
        B = self.B 
        if self.not_increasing:
            B = min(B, 0)
        hat_s = self.initial[1] + (B * (t))
        self.final = (current_t, hat_s)

    def can_add(self, current_t: float, value: float) -> bool:
        n = self.n + 1
        y = value - self.initial[1]
        t = current_t - self.initial[0]
        xx = self.xx + (((t * t) - self.xx)) / n
        xy = self.xy + (((t * y) - self.xy)) / n
        yy = self.yy + (((y * y) - self.yy)) / n
        if xx > 0:
            B = xy / xx
        else:
            B = 0
        SSR = (yy - B * xy) * self.n
        new_segment = False
        if self.n > 15:
            mean_error = np.mean(self.segment_error)
            std_error = np.std(self.segment_error)
            new_segment = SSR > mean_error + 1.5 *std_error
        else:
            new_segment = False
        return not new_segment

    def add(self, current_t: float, value: float):
        self.n = self.n + 1
        y = value - self.initial[1]
        t = current_t - self.initial[0]

        self.xx = self.xx + (((t * t) - self.xx)) / self.n
        self.xy = self.xy + (((t * y) - self.xy)) / self.n
        self.yy = self.yy + (((y * y) - self.yy)) / self.n
        if self.xx > 0:
            self.B = self.xy / self.xx
        else:
            self.B = 0
        SSR = (self.yy - self.B * self.xy) * self.n
        self.segment_error.append(SSR)
        self.compute_endpoint(current_t)


class PiecewesieLinearFunction:
    """Function defined picewise with linear segments

    Parameters
    ----------
    segments: List[Segment]
        List of segments that compose the function
    """

    def __init__(self, segments: List[Segment]):
        self.parameters = []
        self.limits = []
        for s in segments:
            t1, y1 = s.initial
            t2, y2 = s.final
            if (t2 - t1) == 0:
                m = 0
            else:
                m = (y2 - y1) / (t2 - t1)
            b = y1 - m * t1
            self.parameters.append((m, b))
            self.limits.append((t1, t2))


    def predict_line(self, x_values):
        return [self.predict(x) for x in x_values]

    def predict(self, x):
        for i, (l1, l2) in enumerate(self.limits):
            if x >= l1 and x <= l2:
                m, b = self.parameters[i]

                return m * x + b
        m, b = self.parameters[-1]
        return m * x + b

    def zero(self):
        for i, ((l1, l2), (m, b)) in enumerate(zip(self.limits, self.parameters)):
            if m == 0:
                continue
            x_0 = -b / m
            if (x_0 >= l1) and (x_0 <= l2):
                return x_0       
        
        m, b = self.parameters[-1]
        while m == 0:
            return 0
        return -b / m


class PiecewiseLinearRegression:
    """Perform a picewise linear regression

    The method is the one presented in:

    Time and Memory Efficient Online Piecewise Linear Approximation of Sensor Signals
    Florian Grützmacher, Benjamin Beichler, Albert Hein,Dand Thomas Kirste and Christian Haubelt

    """

    def __init__(self, not_increasing:bool=False):
        self.segments = []
        self.not_increasing = not_increasing

    def add_point(
        self,
        t: float,
        s: float,
    ):
        """Add a new point to the regression

        Parameters
        ----------
        t : float
            x component
        s : float
            y component
        """
        if len(self.segments) == 0:
            self.segments.append(Segment((t, s), not_increasing=self.not_increasing))
            return

        if self.segments[-1].can_add(t, s):
            self.segments[-1].add(t, s)
        else:
            self.segments.append(Segment(self.segments[-1].final, not_increasing=self.not_increasing))
            self.segments[-1].add(t, s)

    def finish(self) -> PiecewesieLinearFunction:
        """Complete last unfinished segment and return the model computed

        Returns
        -------
        PiecewesieLinearFunction
            The picewise linear model fitted
        """
        return PiecewesieLinearFunction(self.segments)



import logging
import os
from pathlib import Path
from typing import Callable, List, Optional, Tuple, Union


import pandas as pd
from tqdm.auto import tqdm

from ceruleo.dataset.builder.cycles_splitter import (
    CyclesSplitter,
    FailureDataCycleSplitter,
)
from ceruleo.dataset.builder.output import OutputMode
from ceruleo.dataset.builder.rul_column import RULColumn
from ceruleo.dataset.ts_dataset import PDMDataset

logger = logging.getLogger(__name__)


def load_dataframe(path: Union[str, Path]) -> pd.DataFrame:
    if isinstance(path, str):
        path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"File {path} does not exist")
    if path.suffix == ".csv":
        return pd.read_csv(path)
    if path.suffix == ".parquet":
        return pd.read_parquet(path)
    if path.suffix == ".xlsx":
        return pd.read_excel(path)
    raise ValueError(f"Unsupported file format {path.suffix}")


class DatasetBuilder:
    splitter: CyclesSplitter
    output_mode: OutputMode
    rul_column: RULColumn
    dataframe_loader: Callable[[Union[str, Path]], pd.DataFrame]
    index_column: Optional[str]

    def __init__(
        self,
        dataframe_loader: Callable[[Union[str, Path]], pd.DataFrame] = load_dataframe,
    ):
        """Initializes the builder."""
        self.output_mode = None
        self.splitter = None
        self.dataframe_loader = dataframe_loader
        self.index_column = None
        self.rul_column = None


    @staticmethod
    def two_files_format():
        return DatasetBuilder().set_splitting_method(FailureDataCycleSplitter())

    def set_splitting_method(self, splitter: CyclesSplitter):
        self.splitter = splitter
        return self

    def set_index_column(self, index_column: str):
        self.index_column = index_column
        return self

    def set_machine_id_feature(self, name: str):
        self._machine_type_feature = name
        return self

    def set_rul_column_method(self, rul_column: RULColumn):
        self.rul_column = rul_column
        return self

    def set_output_mode(self, output_mode: OutputMode):
        self.output_mode = output_mode
        return self

    def _validate(self):
        if self.output_mode is None:
            raise ValueError("Output mode not set")
        if self.splitter is None:
            raise ValueError("Splitting method not set")

    def build(self, input_path: Path):
        self._validate()
        self.splitter.split(input_path, self.output_mode)

    def prepare_from_data_fault_pairs_files(
        self, data_fault_pairs: Union[Tuple[str, str], List[Tuple[str, str]]]
    ):
        if not isinstance(data_fault_pairs, list):
            data_fault_pairs = [data_fault_pairs]

        if not isinstance(self.splitter, FailureDataCycleSplitter):
            raise ValueError(
                "This method is only available for FailureDataCycleSplitter"
            )

        common_path_prefix = os.path.commonprefix(
            [data for data, fault in data_fault_pairs]
        )

        for i, (data, fault) in enumerate(tqdm(data_fault_pairs)):
            df_data = self.dataframe_loader(data)
            df_faults = self.dataframe_loader(fault)
            cycles_in_file = self.splitter.split(df_data, df_faults)
            for j, ds in enumerate(cycles_in_file):
                cycle_id = f"{i+1}_{j+1}"
                self._build_and_store_cycle(
                    ds,
                    cycle_id,
                    metadata={
                        "Raw Data Filename": str(data.relative_to(common_path_prefix)),
                        "Raw Fault Filename": str(
                            fault.relative_to(common_path_prefix)
                        ),
                    },
                )
        self.output_mode.finish()

    def prepare_from_single_files(self, file_list: Union[Path, List[Path]]):
        if not isinstance(file_list, list):
            file_list = [file_list]

        common_path_prefix = os.path.commonprefix(
            [data for data in file_list]
        )
        for i, data in enumerate(tqdm(file_list)):
            df_data = self.dataframe_loader(data)
            cycles_in_file = self.splitter.split(df_data,)
            for j, ds in enumerate(cycles_in_file):
                cycle_id = f"{i+1}_{j+1}"
                self._build_and_store_cycle(
                    ds,
                    cycle_id,
                    metadata={
                        "Raw Data Filename": str(data.relative_to(common_path_prefix)),
                    },
                )
        self.output_mode.finish()

    def build_from_single_files(self, file_list: Union[Path, List[Path]]):
        self.prepare_from_single_files(file_list)
        return self.output_mode.build_dataset(self)

    def build_from_data_fault_pairs_files(
        self, data_fault_pairs: Union[Tuple[str, str], List[Tuple[str, str]]]
    ) -> PDMDataset:
        self.prepare_from_data_fault_pairs_files(data_fault_pairs)
        return self.output_mode.build_dataset(self)

    def prepare_from_df(
        self, data: Union[pd.DataFrame, List[pd.DataFrame]]
    ) -> PDMDataset:
        if not isinstance(data, list):
            data = [data]
        self._validate()
        for i, data_element in enumerate(data):
            for j, ds in enumerate(self.splitter.split(data_element)):
                cycle_id = f"{i+1}_{j+1}"
                self._build_and_store_cycle(ds, cycle_id)
            self.output_mode.finish()

    def build_from_df(self, data: Union[pd.DataFrame, List[pd.DataFrame]]):
        self.prepare_from_df(data)
        return self.output_mode.build_dataset(self)

    def prepare_from_data_fault_pair(
        self,
        data_fault_pairs: Union[
            Tuple[pd.DataFrame, pd.DataFrame], List[Tuple[pd.DataFrame, pd.DataFrame]]
        ],
    ):
        if not isinstance(data_fault_pairs, list):
            data_fault_pairs = [data_fault_pairs]

        if not isinstance(self.splitter, FailureDataCycleSplitter):
            raise ValueError(
                "This method is only available for FailureDataCycleSplitter"
            )
        for i, (data, fault) in enumerate(tqdm(data_fault_pairs)):
            cycles_in_file = self.splitter.split(data, fault)
            for j, ds in enumerate(cycles_in_file):
                cycle_id = f"{i+1}_{j+1}"
                self._build_and_store_cycle(
                    ds,
                    cycle_id,
                )
        self.output_mode.finish()

    def build_from_data_fault_pair(
        self,
        data_fault_pairs: Union[
            Tuple[pd.DataFrame, pd.DataFrame], List[Tuple[pd.DataFrame, pd.DataFrame]]
        ],
    ) -> PDMDataset:
        self.prepare_from_data_fault_pair(data_fault_pairs)
        return self.output_mode.build_dataset(self)

    def _build_and_store_cycle(
        self, ds: pd.DataFrame, cycle_id: any, metadata: dict = {}
    ):
        ds["RUL"] = self.rul_column.get(ds)
        if self.index_column is not None:
            ds.set_index(self.index_column, inplace=True)
        self.output_mode.store(f"Cycle_{cycle_id}", ds, metadata)



from copy import deepcopy

import numpy as np
import pandas as pd

from ceruleo.models.model import TrainableModel
from ceruleo.transformation.utils import column_names_window
from sklearn.base import BaseEstimator, TransformerMixin
from tqdm.auto import tqdm


def xgboost_feature_importance(dataset, window, step, transformer):
    from ceruleo.models.xgboost_ import XGBoostModel
    model = XGBoostModel(window, step, transformer)
    model.fit(dataset)
    names = transformer.column_names()
    pd.DataFrame({
        'names': names,
        'importance': model.feature_importances()
    })


class PermuteColumn(BaseEstimator, TransformerMixin):
    def __init__(self, idx: int):
        self.col_idx = idx

    def fit(self, X, y=None):
        return self

    def partial_fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        shuffling_idx = np.arange(X.shape[0])
        np.random.shuffle(shuffling_idx)
        X = X.copy()
        X[:, self.col_idx] = X[shuffling_idx, self.col_idx]
        return X


def permtuation_feature_importance(model: TrainableModel, dataset, metric, n_repeats: int = 5):
    import gc
    y_true = model.true_values(dataset)
    y_pred = model.predict(dataset)
    baseline_score = metric(np.squeeze(y_true), np.squeeze(y_pred))

    scores_dict = {
        'baseline': baseline_score
    }
    original_pipeline = deepcopy(model.transformer.transformerX)
    model.transformer.transformerX.steps.append(
        ('permute', PermuteColumn(-1)))
    for col_idx, column in tqdm(list(enumerate(model.transformer.column_names))):
        scores = np.zeros(n_repeats)
        model.transformer.transformerX.steps[-1][1].col_idx = col_idx
        for n_round in range(n_repeats):
            y_pred = model.predict(dataset)
            score = metric(np.squeeze(y_true), np.squeeze(y_pred))
            scores[n_round] = score
        scores_dict[column] = scores
        gc.collect()
    model.transformer.transformerX = original_pipeline
    return scores_dict


def coefficient_table(model: TrainableModel) -> pd.DataFrame:
    if not hasattr(model.model, 'coef_'):
        raise ValueError('Model does not have coef_')
    cnames = column_names_window(model.transformer.columns(), model.window)

    return pd.DataFrame(zip(cnames, model.model.coef_),
                        columns=['Column', 'Coef'])



from typing import Optional

import numpy as np
import pandas as pd
from ceruleo.transformation import TransformerStep
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_is_fitted
from ceruleo.transformation.features.tdigest import TDigest
from ceruleo.transformation.utils import QuantileEstimator
from sklearn.ensemble import IsolationForest


class IQROutlierRemover(TransformerStep):
    """
    Remove values outside (Q1 - margin*IQR, Q2 + margin*IQR)

    If clip is True the values will be clipped between the range,
    otherwise the values are going to be replaced by inf and -inf

    Parameters:
        lower_quantile: Lower quantile threshold for the non-anomalous values, by feault 0.25
        upper_quantile: Upper quantile threshold for the non-anomalous values, by feault 0.75
        margin: How many times the IQR gets multiplied, by default 0.75
        proportion_to_sample: If you want to compute the quantiles in an smaller proportion of data
            you can specify it,by default 1.0
        clip: Wether to clip the values outside the range, by default False
        name: Name of the step, by default None

    """

    def __init__(
        self,
        lower_quantile: float = 0.25,
        upper_quantile: float = 0.75,
        margin: float=1.5,
        proportion_to_sample: float=1.0,
        clip: bool = False,
        name: Optional[str] = None,
        prefer_partial_fit: bool = False,
    ):

        super().__init__(name=name, prefer_partial_fit=prefer_partial_fit)
        self.margin = margin
        self.proportion_to_sample = proportion_to_sample
        self.tdigest_dict = None
        self.lower_quantile = lower_quantile
        self.upper_quantile = upper_quantile
        self.clip = clip

    def partial_fit(self, X: pd.DataFrame):
        """
        Compute the quantiles of the data and the interquartile range incrementally

        Parameters:
            X: Input life
        """
        if X.shape[0] == 1:
            return self
        if self.proportion_to_sample < 1:
            sampled_points = np.random.choice(
                X.shape[0], int(X.shape[0] * self.proportion_to_sample), replace=False
            )
            X = X.iloc[sampled_points, :]
        if self.tdigest_dict is None:
            self.tdigest_dict = {c: TDigest(100) for c in X.columns}
        for c in X.columns:
            self.tdigest_dict[c] = self.tdigest_dict[c].merge_unsorted(X[c].values)

        self.Q1 = {
            c: self.tdigest_dict[c].estimate_quantile(self.lower_quantile)
            for c in self.tdigest_dict.keys()
        }

        self.Q3 = {
            c: self.tdigest_dict[c].estimate_quantile(self.upper_quantile)
            for c in self.tdigest_dict.keys()
        }

        self.IQR = {c: self.Q3[c] - self.Q1[c] for c in self.Q1.keys()}
        return self

    def fit(self, X: pd.DataFrame):
        """
        Compute the quantiles of the data and the interquartile range incrementally

        Parameters:
            X: Input life
        """
        if self.proportion_to_sample < 1:
            sampled_points = np.random.choice(
                X.shape[0], int(X.shape[0] * self.proportion_to_sample), replace=False
            )
            X = X.iloc[sampled_points, :]
        self.Q1 = X.quantile(self.lower_quantile)
        self.Q3 = X.quantile(self.upper_quantile)
        self.IQR = (self.Q3 - self.Q1).to_dict()
        self.Q1 = self.Q1.to_dict()
        self.Q3 = self.Q3.to_dict()
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Remove the outliers from the input life. 

        Parameters:
            X: Input life
        
        Returns:
            A new DataFrame with the outliers removed
        """
        X = X.copy()
        check_is_fitted(self, "Q1")
        check_is_fitted(self, "Q3")
        check_is_fitted(self, "IQR")
        for c in X.columns:
            min_value = self.Q1[c] - self.margin * self.IQR[c]
            mask = X[c] < min_value
            if not self.clip:
                X.loc[mask, c] = -np.inf
            else:
                X.loc[mask, c] = min_value
            max_value = self.Q3[c] + self.margin * self.IQR[c]
            mask = X[c] > (max_value)
            if not self.clip:
                X.loc[mask, c] = np.inf
            else:
                X.loc[mask, c] = max_value
        return X

    def description(self):
        name = super().description()
        data = []
        for k in self.Q1.keys():
            data.append((k, {"Q1": self.Q1[k], "Q3": self.Q3[k], "IQR": self.IQR[k]}))
        return (name, data)


class BeyondQuartileOutlierRemover(TransformerStep):
    """
    Remove values outside (Q1, Q3)

    If clip is True the values will be clipped between the range,
    otherwise the values are going to be replaced by inf and -inf

    Parameters:
        lower_quantile:  Lower quantile threshold for the non-anomalous values, by default 0.25
        upper_quantile: Upper quantile threshold for the non-anomalous values, by default 0.75
        clip: Wether to clip the values outside the range, by default False
        name: Name of the step, by default None
    """

    def __init__(
        self,
        lower_quantile: float = 0.25,
        upper_quantile: float = 0.75,
        subsample: float = 1.0,
        clip: bool = False,
        name: Optional[str] = None,
        prefer_partial_fit:bool=False
    ):

        super().__init__(name=name, prefer_partial_fit=prefer_partial_fit)
        self.tdigest_dict = None
        self.lower_quantile = lower_quantile
        self.upper_quantile = upper_quantile
        self.clip = clip
        self.subsample= subsample
        self.Q1 = None
        self.Q3 = None
        self.quantile_estimator = None

    def partial_fit(self, X: pd.DataFrame):
        """
        Compute the quantiles of the data incrementally

        Parameters:
            X: Input life
        """
        if X.shape[0] == 1:
            return self
        if self.quantile_estimator is None:
            self.quantile_estimator = QuantileEstimator(
               tdigest_size=100, subsample=self.subsample
            )

        self.quantile_estimator.update(X.select_dtypes(include="number"))
        return self

    def fit(self, X: pd.DataFrame):
        """
        Compute the quantiles of the data

        Parameters:
            X: Input life
        """
        if self.subsample < 1:

            sampled_points = np.random.choice(
                X.shape[0], int(X.shape[0] * self.subsample), replace=False
            )
            X = X.iloc[sampled_points, :]
        self.Q1 = X.quantile(self.lower_quantile)
        self.Q3 = X.quantile(self.upper_quantile)

        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """ 
        Remove the outliers from the input life.

        Parameters:
            X: Input life
        
        Returns:
            A new DataFrame with the outliers removed
        """

        if self.Q1 is None:
            self.Q1 = self.quantile_estimator.estimate_quantile(self.lower_quantile)
            self.Q3 = self.quantile_estimator.estimate_quantile(self.upper_quantile)

        new_X = X.copy()
        

        if self.clip:
            new_X.clip(lower=self.Q1, upper=self.Q3, inplace=True, axis=1)
        else:
            new_X[new_X < self.Q1] = -np.inf
            new_X[new_X > self.Q3] = np.inf            
        return new_X

    def description(self):
        name = super().description()
        data = []
        for k in self.Q1.keys():
            data.append((k, {"Q1": self.Q1[k], "Q3": self.Q3[k], "IQR": self.IQR[k]}))
        return (name, data)


class ZScoreOutlierRemover(TransformerStep):
    """
    Remove values outside (mean - number_of_std_allowed*std, mean + number_of_std_allowed*std). The outliers are set to NaN

    Parameters:
        number_of_std_allowed: Number of standard deviations to consider a point an outlier
        name: Name of the step, by default None
    """
    #X = np.random.rand(500, 5) * np.random.randn(500, 5) * 15
    #imput = ZScoreImputer(1.5)
    #imput.fit(X)
    #X_t = imput.transform(X)

    def __init__(
        self,
        *,
        number_of_std_allowed,
        name: str = None,
        prefer_partial_fit: bool = False,
    ):
        super().__init__(name=name, prefer_partial_fit=prefer_partial_fit)
        self.number_of_std_allowed = number_of_std_allowed
        self.scaler = StandardScaler()

    def fit(self, X: pd.DataFrame):
        """
        Fit a StandardScaler to the data

        Parameters:
            X: Input life
        """
        self.scaler.fit(X)
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Remove the outliers from the input life.

        Parameters:
            X: Input life

        Returns:
            A new DataFrame with the outliers removed
        """
        X_new = self.scaler.transform(X)
        X_new[np.abs(X_new) > self.number_of_std_allowed] = np.nan
        return pd.DataFrame(X_new, columns=X.columns, index=X.index)


class EWMAOutOfRange(TransformerStep):
    """
    Compute the EWMA limits  and mark as NaN points outside UCL and LCL

    Parameters:
        lambda_: Parameter for the EWMA, by default 0.5
        return_mask: Wether to return a mask with the outliers or the original data with the outliers marked as NaN, by default False
        name: Name of the step, by default None
    """

    def __init__(
        self,
        *,
        lambda_ : float=0.5,
        return_mask: bool = False,
        name: Optional[str] = None,
        prefer_partial_fit: bool = False,
    ):
        super().__init__(name=name, prefer_partial_fit=prefer_partial_fit)
        self.lambda_ = lambda_
        self.UCL = None
        self.LCL = None
        self.columns = None
        self.return_mask = return_mask

    def partial_fit(self, X: pd.DataFrame, y=None):
        """
        Compute the EWMA limits incrementally

        Parameters:
            X: Input life
        """
        if self.columns is None:
            self.columns = X.columns.values
        else:
            self.columns = [c for c in self.columns if c in X.columns]
        if self.LCL is not None:
            self.LCL = self.LCL.loc[self.columns].copy()
            self.UCL = self.UCL.loc[self.columns].copy()
        LCL, UCL = self._compute_limits(X[self.columns].copy())
        self.LCL = np.minimum(LCL, self.LCL) if self.LCL is not None else LCL
        self.UCL = np.maximum(UCL, self.UCL) if self.UCL is not None else UCL
        return self

    def _compute_limits(self, X):

        mean = np.nanmean(X, axis=0)
        s = np.sqrt(self.lambda_ / (2 - self.lambda_)) * np.nanstd(X, axis=0)
        UCL = mean + 3 * s
        LCL = mean - 3 * s
        return (pd.Series(LCL, index=self.columns), pd.Series(UCL, index=self.columns))

    def fit(self, X: pd.DataFrame, y=None):
        """
        Compute the EWMA limits

        Parameters:
            X: Input life
        """
        self.columns = X.columns
        LCL, UCL = self._compute_limits(X)
        self.LCL = LCL
        self.UCL = UCL
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Remove the outliers from the input life.

        Parameters:
            X: Input life
        
        Returns:
            A new DataFrame with the outliers removed
        """
        mask = (X[self.columns] < (self.LCL)) | (X[self.columns] > (self.UCL))
        if self.return_mask:
            return mask.astype("int")
        else:
            X = X.copy()
            X[mask] = np.nan
            return X

# We can do the same with Rolling Median

class RollingMeanOutlierRemover(TransformerStep):
    """
    Compute the rolling mean and use it to compute the upper and lower bound to define outliers 

    Parameters:
        window: Window for the rolling mean, by default 15
        lambda_: Multiplier of the std used to define the bounds, by default 3
        return_mask: Wether to return a mask with the outliers or the original data with the outliers marked as NaN, by default False
        name: Name of the step, by default None
    """

    def __init__(
        self,
        *,
        window: int = 15,
        lambda_: float = 3,
        return_mask: bool = False,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self.window = window
        self.lambda_ = lambda_
        self.return_mask = return_mask

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Remove the outliers from the input life.

        Parameters:
            X: Input life
        
        Returns:
            A new DataFrame with the outliers removed
        """
        r = X.rolling(self.window, min_periods=1)
        std = r.quantile(0.75) -  r.quantile(0.25)
        upper = r.median() + (self.lambda_ * std)
        lower = r.median() - (self.lambda_ * std)
        mask = (X > upper) | (X < lower)
        if self.return_mask:
            return mask.astype("int")
        else:
            X = X.copy()
            X[(X > upper)] = np.minimum(upper.values, X) 
            X[(X < upper)] = np.maximum(lower.values, X) 

            #X[mask] = np.nan
            return X


class IsolationForestOutlierRemover(TransformerStep):
    """ 
    Remove outliers using Isolation Forests to detect them.

    Parameters:
        n_estimators: Number of trees in the forest, by default 100
        name: Name of the step, by default None
    """
    def __init__(self, *, n_estimators=100, **kwargs):
        super().__init__(prefer_partial_fit=False, **kwargs)
        self.n_estimators = n_estimators
        self.forests = {}

    def fit(self, X: pd.DataFrame):
        """
        Fit the Isolation Forest model to the data

        Parameters:
            X: Input life
        """
        for c in X.columns:
            self.forests[c] = IsolationForest(n_estimators=self.n_estimators).fit(X[c].values.reshape(-1, 1) )
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Remove the outliers from the input life.
        
        Parameters:
            X: Input life
        
        Returns:
            A new DataFrame with the outliers removed
        """
        X_new = X.copy()
        for c in X.columns:
            r = self.forests[c].predict(X[c].values.reshape(-1, 1) )
            X_new.loc[r == -1, c] = np.nan
        return X_new
